# sage.doctest: needs sage.flint
"""
Benchmarks for matrices

This file has many functions for computing timing benchmarks
of various methods for random matrices with given bounds for
the entries.  The systems supported are Sage and Magma.

The basic command syntax is as follows::

    sage: import sage.matrix.benchmark as b
    sage: print("starting"); import sys; sys.stdout.flush(); b.report([b.det_ZZ], 'Test', systems=['sage'])
    starting...
    ======================================================================
              Test
    ======================================================================
    ...
    ======================================================================
"""

import sys

from sage.matrix.constructor import Matrix, random_matrix
from sage.misc.lazy_import import lazy_import
from sage.misc.timing import cputime
from sage.rings.finite_rings.finite_field_constructor import FiniteField as GF
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ

if sys.platform != 'win32':
    from cysignals.alarm import AlarmInterrupt, alarm, cancel_alarm

lazy_import('sage.interfaces.magma', 'magma')

verbose = False

timeout = 60


def report(F, title, systems=['sage', 'magma'], **kwds):
    """
    Run benchmarks with default arguments for each function in the list F.

    INPUT:

    - ``F`` -- list of callables used for benchmarking
    - ``title`` -- string describing this report
    - ``systems`` -- list of systems (supported entries are 'sage' and 'magma')
    - ``**kwds`` -- keyword arguments passed to all functions in ``F``

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: print("starting"); import sys; sys.stdout.flush(); b.report([b.det_ZZ], 'Test', systems=['sage'])
        starting...
        ======================================================================
                  Test
        ======================================================================
        ...
        ======================================================================
    """
    import os
    if len(systems) > 2:
        raise NotImplementedError("at most two systems ('sage' or 'magma')")
    print('=' * 70)
    print(' ' * 10 + title)
    print('=' * 70)
    os.system('uname -a')
    print('\n')
    for f in F:
        print("-"*70)
        print(f.__doc__.strip())
        print(('%15s' * len(systems)) % tuple(systems))
        w = []
        for s in systems:
            alarm(timeout)
            try:
                t = f(system=s, **kwds)
            except AlarmInterrupt:
                t = -timeout
            cancel_alarm()
            w.append(float(t))
        if len(w) > 1:
            if w[1] == 0:
                w.append(0.0)
            else:
                w.append(w[0]/w[1])

        w = tuple(w)
        print(('%15.3f'*len(w)) % w)
    print('=' * 70)


#######################################################################
# Dense Benchmarks over ZZ
#######################################################################

def report_ZZ(**kwds):
    """
    Reports all the benchmarks for integer matrices and few
    rational matrices.

    INPUT:

    - ``**kwds`` -- passed through to :func:`report`

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: print("starting"); import sys; sys.stdout.flush(); b.report_ZZ(systems=['sage'])  # long time (15s on sage.math, 2012)
        starting...
        ======================================================================
        Dense benchmarks over ZZ
        ======================================================================
        ...
        ======================================================================
    """
    F = [vecmat_ZZ, rank_ZZ, rank2_ZZ, charpoly_ZZ, smithform_ZZ,
         det_ZZ, det_QQ, matrix_multiply_ZZ, matrix_add_ZZ,
         matrix_add_ZZ_2,
         nullspace_ZZ]

    title = 'Dense benchmarks over ZZ'
    report(F, title, **kwds)

# Integer Nullspace


def nullspace_ZZ(n=200, min=0, max=2**32, system='sage'):
    """
    Nullspace over ZZ:
    Given a n+1 x n matrix over ZZ with random entries
    between min and max, compute the nullspace.

    INPUT:

    - ``n`` -- matrix dimension (default: ``200``)
    - ``min`` -- minimal value for entries of matrix (default: ``0``)
    - ``max`` -- maximal value for entries of matrix (default: ``2**32``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.nullspace_ZZ(200)
        sage: tm = b.nullspace_ZZ(200, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(ZZ, n+1, n, x=min, y=max+1).change_ring(QQ)
        t = cputime()
        v = A.kernel()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := RMatrixSpace(RationalField(), n+1,n)![Random(%s,%s) : i in [1..n*(n+1)]];
t := Cputime();
K := Kernel(A);
s := Cputime(t);
""" % (n, min, max)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)


def charpoly_ZZ(n=100, min=0, max=9, system='sage'):
    """
    Characteristic polynomial over ZZ:
    Given a n x n matrix over ZZ with random entries between min and
    max, compute the charpoly.

    INPUT:

    - ``n`` -- matrix dimension (default: ``100``)
    - ``min`` -- minimal value for entries of matrix (default: ``0``)
    - ``max`` -- maximal value for entries of matrix (default: ``9``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.charpoly_ZZ(100)
        sage: tm = b.charpoly_ZZ(100, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(ZZ, n, n, x=min, y=max+1)
        t = cputime()
        v = A.charpoly()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := MatrixAlgebra(IntegerRing(), n)![Random(%s,%s) : i in [1..n^2]];
t := Cputime();
K := CharacteristicPolynomial(A);
s := Cputime(t);
""" % (n, min, max)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)


def rank_ZZ(n=700, min=0, max=9, system='sage'):
    """
    Rank over ZZ:
    Given a n x (n+10) matrix over ZZ with random entries
    between min and max, compute the rank.

    INPUT:

    - ``n`` -- matrix dimension (default: ``700``)
    - ``min`` -- minimal value for entries of matrix (default: ``0``)
    - ``max`` -- maximal value for entries of matrix (default: ``9``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.rank_ZZ(300)
        sage: tm = b.rank_ZZ(300, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(ZZ, n, n+10, x=min, y=max+1)
        t = cputime()
        v = A.rank()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := RMatrixSpace(IntegerRing(), n, n+10)![Random(%s,%s) : i in [1..n*(n+10)]];
t := Cputime();
K := Rank(A);
s := Cputime(t);
""" % (n, min, max)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)


def rank2_ZZ(n=400, min=0, max=2**64, system='sage'):
    """
    Rank 2 over ZZ:
    Given a (n + 10) x n matrix over ZZ with random entries
    between min and max, compute the rank.

    INPUT:

    - ``n`` -- matrix dimension (default: ``400``)
    - ``min`` -- minimal value for entries of matrix (default: ``0``)
    - ``max`` -- maximal value for entries of matrix (default: ``2**64``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.rank2_ZZ(300)
        sage: tm = b.rank2_ZZ(300, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(ZZ, n+10, n, x=min, y=max+1)
        t = cputime()
        v = A.rank()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := RMatrixSpace(IntegerRing(), n+10, n)![Random(%s,%s) : i in [1..n*(n+10)]];
t := Cputime();
K := Rank(A);
s := Cputime(t);
""" % (n, min, max)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)

# Smith Form


def smithform_ZZ(n=128, min=0, max=9, system='sage'):
    """
    Smith Form over ZZ:
    Given a n x n matrix over ZZ with random entries
    between min and max, compute the Smith normal form.

    INPUT:

    - ``n`` -- matrix dimension (default: ``128``)
    - ``min`` -- minimal value for entries of matrix (default: ``0``)
    - ``max`` -- maximal value for entries of matrix (default: ``9``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.smithform_ZZ(100)
        sage: tm = b.smithform_ZZ(100, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(ZZ, n, n, x=min, y=max+1)
        t = cputime()
        v = A.elementary_divisors()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := MatrixAlgebra(IntegerRing(), n)![Random(%s,%s) : i in [1..n^2]];
t := Cputime();
K := ElementaryDivisors(A);
s := Cputime(t);
""" % (n, min, max)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)


def matrix_multiply_ZZ(n=300, min=-9, max=9, system='sage', times=1):
    """
    Matrix multiplication over ZZ
    Given an n x n matrix A over ZZ with random entries
    between min and max, inclusive, compute A * (A+1).

    INPUT:

    - ``n`` -- matrix dimension (default: ``300``)
    - ``min`` -- minimal value for entries of matrix (default: ``-9``)
    - ``max`` -- maximal value for entries of matrix (default: ``9``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)
    - ``times`` -- number of experiments (default: ``1``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.matrix_multiply_ZZ(200)
        sage: tm = b.matrix_multiply_ZZ(200, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(ZZ, n, n, x=min, y=max+1)
        B = A + 1
        t = cputime()
        for z in range(times):
            v = A * B
        return cputime(t)/times
    if system == 'magma':
        code = """
n := %s;
A := MatrixAlgebra(IntegerRing(), n)![Random(%s,%s) : i in [1..n^2]];
B := A + 1;
t := Cputime();
for z in [1..%s] do
    K := A * B;
end for;
s := Cputime(t);
""" % (n, min, max, times)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))/times
    raise ValueError('unknown system "%s"' % system)


def matrix_add_ZZ(n=200, min=-9, max=9, system='sage', times=50):
    """
    Matrix addition over ZZ
    Given an n x n matrix A and B over ZZ with random entries between
    ``min`` and ``max``, inclusive, compute A + B ``times`` times.

    INPUT:

    - ``n`` -- matrix dimension (default: ``200``)
    - ``min`` -- minimal value for entries of matrix (default: ``-9``)
    - ``max`` -- maximal value for entries of matrix (default: ``9``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)
    - ``times`` -- number of experiments (default: ``50``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.matrix_add_ZZ(200)
        sage: tm = b.matrix_add_ZZ(200, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(ZZ, n, n, x=min, y=max+1)
        B = random_matrix(ZZ, n, n, x=min, y=max+1)
        t = cputime()
        for z in range(times):
            v = A + B
        return cputime(t)/times
    if system == 'magma':
        code = """
n := %s;
min := %s;
max := %s;
A := MatrixAlgebra(IntegerRing(), n)![Random(min,max) : i in [1..n^2]];
B := MatrixAlgebra(IntegerRing(), n)![Random(min,max) : i in [1..n^2]];
t := Cputime();
for z in [1..%s] do
    K := A + B;
end for;
s := Cputime(t);
""" % (n, min, max, times)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))/times
    raise ValueError('unknown system "%s"' % system)


def matrix_add_ZZ_2(n=200, bits=16, system='sage', times=50):
    """
    Matrix addition over ZZ.
    Given an n x n matrix A and B over ZZ with random ``bits``-bit
    entries, compute A + B.

    INPUT:

    - ``n`` -- matrix dimension (default: ``200``)
    - ``bits`` -- bitsize of entries
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)
    - ``times`` -- number of experiments (default: ``50``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.matrix_add_ZZ_2(200)
        sage: tm = b.matrix_add_ZZ_2(200, system='magma')  # optional - magma
    """
    b = 2**bits
    return matrix_add_ZZ(n=n, min=-b, max=b,system=system, times=times)


def det_ZZ(n=200, min=1, max=100, system='sage'):
    """
    Dense integer determinant over ZZ.
    Given an n x n matrix A over ZZ with random entries
    between min and max, inclusive, compute det(A).

    INPUT:

    - ``n`` -- matrix dimension (default: ``200``)
    - ``min`` -- minimal value for entries of matrix (default: ``1``)
    - ``max`` -- maximal value for entries of matrix (default: ``100``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.det_ZZ(200)
        sage: tm = b.det_ZZ(200, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(ZZ, n, n, x=min, y=max+1)
        t = cputime()
        d = A.determinant()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := MatrixAlgebra(IntegerRing(), n)![Random(%s,%s) : i in [1..n^2]];
t := Cputime();
d := Determinant(A);
s := Cputime(t);
""" % (n, min, max)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)


def det_QQ(n=300, num_bound=10, den_bound=10, system='sage'):
    """
    Dense rational determinant over QQ.
    Given an n x n matrix A over QQ with random entries
    with numerator bound and denominator bound, compute det(A).

    INPUT:

    - ``n`` -- matrix dimension (default: ``200``)
    - ``num_bound`` -- numerator bound, inclusive (default: ``10``)
    - ``den_bound`` -- denominator bound, inclusive (default: ``10``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.det_QQ(200)
        sage: ts = b.det_QQ(10, num_bound=100000, den_bound=10000)
        sage: tm = b.det_QQ(200, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(QQ, n, n, num_bound=num_bound, den_bound=den_bound)
        t = cputime()
        d = A.determinant()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := MatrixAlgebra(RationalField(), n)![Random(%s,%s)/Random(1,%s) : i in [1..n^2]];
t := Cputime();
d := Determinant(A);
s := Cputime(t);
""" % (n,-num_bound, num_bound, den_bound)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)


def vecmat_ZZ(n=300, min=-9, max=9, system='sage', times=200):
    """
    Vector matrix multiplication over ZZ.

    Given an n x n  matrix A over ZZ with random entries
    between min and max, inclusive, and v the first row of A,
    compute the product v * A.

    INPUT:

    - ``n`` -- matrix dimension (default: ``300``)
    - ``min`` -- minimal value for entries of matrix (default: ``-9``)
    - ``max`` -- maximal value for entries of matrix (default: ``9``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)
    - ``times`` -- number of runs (default: ``200``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.vecmat_ZZ(300)  # long time
        sage: tm = b.vecmat_ZZ(300, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(ZZ, n, n, x=min, y=max+1)
        v = A.row(0)
        t = cputime()
        for z in range(times):
            w = v * A
        return cputime(t)/times
    if system == 'magma':
        code = """
n := %s;
A := MatrixAlgebra(IntegerRing(), n)![Random(%s,%s) : i in [1..n^2]];
v := A[1];
t := Cputime();
for z in [1..%s] do
    K := v * A;
end for;
s := Cputime(t);
""" % (n, min, max, times)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))/times
    raise ValueError('unknown system "%s"' % system)


#######################################################################
# Dense Benchmarks over GF(p), for small p.
#######################################################################

def report_GF(p=16411, **kwds):
    """
    Run all the reports for finite field matrix operations, for
    prime p=16411.

    INPUT:

    - ``p`` -- ignored
    - ``**kwds`` -- passed through to :func:`report`

    .. NOTE::

        right now, even though p is an input, it is being ignored!  If
        you need to check the performance for other primes, you can
        call individual benchmark functions.

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: print("starting"); import sys; sys.stdout.flush(); b.report_GF(systems=['sage'])
        starting...
        ======================================================================
        Dense benchmarks over GF with prime 16411
        ======================================================================
        ...
        ======================================================================
    """
    F = [rank_GF, rank2_GF, nullspace_GF, charpoly_GF,
         matrix_multiply_GF, det_GF]
    title = 'Dense benchmarks over GF with prime %i' % p
    report(F, title, **kwds)

# Nullspace over GF


def nullspace_GF(n=300, p=16411, system='sage'):
    """
    Given a n+1 x n  matrix over GF(p) with random
    entries, compute the nullspace.

    INPUT:

    - ``n`` -- matrix dimension (default: 300)
    - ``p`` -- prime number (default: ``16411``)
    - ``system`` -- either 'magma' or 'sage' (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.nullspace_GF(300)
        sage: tm = b.nullspace_GF(300, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(GF(p), n, n+1)
        t = cputime()
        v = A.kernel()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := Random(RMatrixSpace(GF(%s), n, n+1));
t := Cputime();
K := Kernel(A);
s := Cputime(t);
""" % (n,p)
        if verbose:
            print(code)
        magma.eval(code)
        return magma.eval('s')
    raise ValueError('unknown system "%s"' % system)


# Characteristic Polynomial over GF

def charpoly_GF(n=100, p=16411, system='sage'):
    """
    Given a n x n matrix over GF with random entries, compute the
    charpoly.

    INPUT:

    - ``n`` -- matrix dimension (default: 100)
    - ``p`` -- prime number (default: ``16411``)
    - ``system`` -- either 'magma' or 'sage' (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.charpoly_GF(100)
        sage: tm = b.charpoly_GF(100, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(GF(p), n, n)
        t = cputime()
        v = A.charpoly()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := Random(MatrixAlgebra(GF(%s), n));
t := Cputime();
K := CharacteristicPolynomial(A);
s := Cputime(t);
""" % (n,p)
        if verbose:
            print(code)
        magma.eval(code)
        return magma.eval('s')
    raise ValueError('unknown system "%s"' % system)


def matrix_add_GF(n=1000, p=16411, system='sage', times=100):
    """
    Given two n x n matrix over GF(p) with random entries, add them.

    INPUT:

    - ``n`` -- matrix dimension (default: 300)
    - ``p`` -- prime number (default: ``16411``)
    - ``system`` -- either 'magma' or 'sage' (default: ``'sage'``)
    - ``times`` -- number of experiments (default: ``100``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.matrix_add_GF(500, p=19)
        sage: tm = b.matrix_add_GF(500, p=19, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(GF(p), n, n)
        B = random_matrix(GF(p), n, n)
        t = cputime()
        for n in range(times):
            v = A + B
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := Random(MatrixAlgebra(GF(%s), n));
B := Random(MatrixAlgebra(GF(%s), n));
t := Cputime();
for z in [1..%s] do
    K := A + B;
end for;
s := Cputime(t);
""" % (n,p,p,times)
        if verbose:
            print(code)
        magma.eval(code)
        return magma.eval('s')
    raise ValueError('unknown system "%s"' % system)


# Matrix multiplication over GF(p)

def matrix_multiply_GF(n=100, p=16411, system='sage', times=3):
    """
    Given an n x n matrix A over GF(p) with random entries, compute
    A * (A+1).

    INPUT:

    - ``n`` -- matrix dimension (default: 100)
    - ``p`` -- prime number (default: ``16411``)
    - ``system`` -- either 'magma' or 'sage' (default: ``'sage'``)
    - ``times`` -- number of experiments (default: ``3``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.matrix_multiply_GF(100, p=19)
        sage: tm = b.matrix_multiply_GF(100, p=19, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(GF(p), n)
        B = A + 1
        t = cputime()
        for n in range(times):
            v = A * B
        return cputime(t) / times
    if system == 'magma':
        code = """
n := %s;
A := Random(MatrixAlgebra(GF(%s), n));
B := A + 1;
t := Cputime();
for z in [1..%s] do
    K := A * B;
end for;
s := Cputime(t);
""" % (n,p,times)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))/times
    raise ValueError('unknown system "%s"' % system)


def rank_GF(n=500, p=16411, system='sage'):
    """
    Rank over GF(p):
    Given a n x (n+10) matrix over GF(p) with random entries, compute the rank.

    INPUT:

    - ``n`` -- matrix dimension (default: 300)
    - ``p`` -- prime number (default: ``16411``)
    - ``system`` -- either 'magma' or 'sage' (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.rank_GF(1000)
        sage: tm = b.rank_GF(1000, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(GF(p), n, n+10)
        t = cputime()
        v = A.rank()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := Random(MatrixAlgebra(GF(%s), n));
t := Cputime();
K := Rank(A);
s := Cputime(t);
""" % (n,p)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)


def rank2_GF(n=500, p=16411, system='sage'):
    """
    Rank over GF(p): Given a (n + 10) x n matrix over GF(p) with
    random entries, compute the rank.

    INPUT:

    - ``n`` -- matrix dimension (default: 300)
    - ``p`` -- prime number (default: ``16411``)
    - ``system`` -- either 'magma' or 'sage' (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.rank2_GF(500)
        sage: tm = b.rank2_GF(500, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(GF(p), n+10, n)
        t = cputime()
        v = A.rank()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := Random(MatrixAlgebra(GF(%s), n));
t := Cputime();
K := Rank(A);
s := Cputime(t);
""" % (n,p)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)


def det_GF(n=400, p=16411 , system='sage'):
    """
    Dense determinant over GF(p).
    Given an n x n matrix A over GF with random entries compute
    det(A).

    INPUT:

    - ``n`` -- matrix dimension (default: 300)
    - ``p`` -- prime number (default: ``16411``)
    - ``system`` -- either 'magma' or 'sage' (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.det_GF(1000)
        sage: tm = b.det_GF(1000, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(GF(p), n, n)
        t = cputime()
        d = A.determinant()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := Random(MatrixAlgebra(GF(%s), n));
t := Cputime();
d := Determinant(A);
s := Cputime(t);
""" % (n,p)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)


#######################################################################
# Dense Benchmarks over QQ
#######################################################################

def hilbert_matrix(n):
    """
    Return the Hilbert matrix of size n over rationals.

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: b.hilbert_matrix(3)
        [  1 1/2 1/3]
        [1/2 1/3 1/4]
        [1/3 1/4 1/5]
    """
    A = Matrix(QQ,n,n)
    for i in range(A.nrows()):
        for j in range(A.ncols()):
            A[i,j] = QQ(1)/((i+1)+(j+1)-1)
    return A

# Reduced row echelon form over QQ


def echelon_QQ(n=100, min=0, max=9, system='sage'):
    """
    Given a n x (2*n) matrix over QQ with random integer entries
    between min and max, compute the reduced row echelon form.

    INPUT:

    - ``n`` -- matrix dimension (default: ``300``)
    - ``min`` -- minimal value for entries of matrix (default: ``-9``)
    - ``max`` -- maximal value for entries of matrix (default: ``9``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.echelon_QQ(100)
        sage: tm = b.echelon_QQ(100, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(ZZ, n, 2*n, x=min, y=max+1).change_ring(QQ)
        t = cputime()
        v = A.echelon_form()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := RMatrixSpace(RationalField(), n, 2*n)![Random(%s,%s) : i in [1..n*2*n]];
t := Cputime();
K := EchelonForm(A);
s := Cputime(t);
""" % (n, min, max)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)

# Invert a matrix over QQ.


def inverse_QQ(n=100, min=0, max=9, system='sage'):
    """
    Given a n x n matrix over QQ with random integer entries
    between min and max, compute the reduced row echelon form.

    INPUT:

    - ``n`` -- matrix dimension (default: ``300``)
    - ``min`` -- minimal value for entries of matrix (default: ``-9``)
    - ``max`` -- maximal value for entries of matrix (default: ``9``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.inverse_QQ(100)
        sage: tm = b.inverse_QQ(100, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(ZZ, n, n, x=min, y=max+1).change_ring(QQ)
        t = cputime()
        v = ~A
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := MatrixAlgebra(RationalField(), n)![Random(%s,%s) : i in [1..n*n]];
t := Cputime();
K := A^(-1);
s := Cputime(t);
""" % (n, min, max)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)


# Matrix multiplication over QQ
def matrix_multiply_QQ(n=100, bnd=2, system='sage', times=1):
    """
    Given an n x n matrix A over QQ with random entries
    whose numerators and denominators are bounded by bnd,
    compute A * (A+1).

    INPUT:

    - ``n`` -- matrix dimension (default: ``300``)
    - ``bnd`` -- numerator and denominator bound (default: ``bnd``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)
    - ``times`` -- number of experiments (default: ``1``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.matrix_multiply_QQ(100)
        sage: tm = b.matrix_multiply_QQ(100, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = random_matrix(QQ, n, n, num_bound=bnd, den_bound=bnd)
        B = A + 1
        t = cputime()
        for z in range(times):
            v = A * B
        return cputime(t)/times
    if system == 'magma':
        A = magma(random_matrix(QQ, n, n, num_bound=bnd, den_bound=bnd))
        code = """
n := %s;
A := %s;
B := A + 1;
t := Cputime();
for z in [1..%s] do
    K := A * B;
end for;
s := Cputime(t);
""" % (n, A.name(), times)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))/times
    raise ValueError('unknown system "%s"' % system)


# Determinant of Hilbert matrix
def det_hilbert_QQ(n=80, system='sage'):
    """
    Run the benchmark for calculating the determinant of the hilbert
    matrix over rationals of dimension n.

    INPUT:

    - ``n`` -- matrix dimension (default: ``300``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.det_hilbert_QQ(50)
        sage: tm = b.det_hilbert_QQ(50, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = hilbert_matrix(n)
        t = cputime()
        d = A.determinant()
        return cputime(t)
    if system == 'magma':
        code = """
h := HilbertMatrix(%s);
tinit := Cputime();
d := Determinant(h);
s := Cputime(tinit);
delete h;
""" % n
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))

# inverse of Hilbert matrix


def invert_hilbert_QQ(n=40, system='sage'):
    """
    Run the benchmark for calculating the inverse of the hilbert
    matrix over rationals of dimension n.

    INPUT:

    - ``n`` -- matrix dimension (default: ``300``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.invert_hilbert_QQ(30)
        sage: tm = b.invert_hilbert_QQ(30, system='magma')  # optional - magma
    """
    if system == 'sage':
        A = hilbert_matrix(n)
        t = cputime()
        d = A**(-1)
        return cputime(t)
    if system == 'magma':
        code = """
h := HilbertMatrix(%s);
tinit := Cputime();
d := h^(-1);
s := Cputime(tinit);
delete h;
""" % n
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))


def MatrixVector_QQ(n=1000, h=100, system='sage', times=1):
    """
    Compute product of square ``n`` matrix by random vector with num and
    denom bounded by ``h`` the given number of ``times``.

    INPUT:

    - ``n`` -- matrix dimension (default: ``300``)
    - ``h`` -- numerator and denominator bound (default: ``bnd``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)
    - ``times`` -- number of experiments (default: ``1``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.MatrixVector_QQ(500)
        sage: tm = b.MatrixVector_QQ(500, system='magma')  # optional - magma
    """
    if system == 'sage':
        V = QQ**n
        v = V.random_element(h)
        M = random_matrix(QQ,n)
        t = cputime()
        for i in range(times):
            w = M*v
        return cputime(t)
    if system == 'magma':
        code = """
            n:=%s;
            h:=%s;
            times:=%s;
            v:=VectorSpace(RationalField(),n)![Random(h)/(Random(h)+1) : i in [1..n]];
            M:=MatrixAlgebra(RationalField(),n)![Random(h)/(Random(h)+1) : i in [1..n^2]];
            t := Cputime();
            for z in [1..times] do
                W:=v*M;
            end for;
            s := Cputime(t);
        """ % (n,h,times)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)


#######################################################################
# Dense Benchmarks over machine reals
# Note that the precision in reals for MAGMA is base 10, while in
# sage it is in base 2
#######################################################################

# Real Nullspace

def nullspace_RR(n=300, min=0, max=10, system='sage'):
    """
    Nullspace over RR:
    Given a n+1 x n matrix over RR with random entries
    between min and max, compute the nullspace.

    INPUT:

    - ``n`` -- matrix dimension (default: ``300``)
    - ``min`` -- minimal value for entries of matrix (default: ``0``)
    - ``max`` -- maximal value for entries of matrix (default: ``10``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.nullspace_RR(100)
        sage: tm = b.nullspace_RR(100, system='magma')  # optional - magma
    """
    if system == 'sage':
        from sage.rings.real_mpfr import RR
        A = random_matrix(ZZ, n+1, n, x=min, y=max+1).change_ring(RR)
        t = cputime()
        v = A.kernel()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := RMatrixSpace(RealField(16), n+1,n)![Random(%s,%s) : i in [1..n*(n+1)]];
t := Cputime();
K := Kernel(A);
s := Cputime(t);
""" % (n, min, max)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)


def nullspace_RDF(n=300, min=0, max=10, system='sage'):
    """
    Nullspace over RDF:
    Given a n+1 x n  matrix over RDF with random entries
    between min and max, compute the nullspace.

    INPUT:

    - ``n`` -- matrix dimension (default: ``300``)
    - ``min`` -- minimal value for entries of matrix (default: ``0``)
    - ``max`` -- maximal value for entries of matrix (default: ``10``)
    - ``system`` -- either ``'sage'`` or ``'magma'`` (default: ``'sage'``)

    EXAMPLES::

        sage: import sage.matrix.benchmark as b
        sage: ts = b.nullspace_RDF(100)  # long time
        sage: tm = b.nullspace_RDF(100, system='magma')  # optional - magma
    """
    if system == 'sage':
        from sage.rings.real_double import RDF
        A = random_matrix(ZZ, n+1, n, x=min, y=max+1).change_ring(RDF)
        t = cputime()
        v = A.kernel()
        return cputime(t)
    if system == 'magma':
        code = """
n := %s;
A := RMatrixSpace(RealField(16), n+1,n)![Random(%s,%s) : i in [1..n*(n+1)]];
t := Cputime();
K := Kernel(A);
s := Cputime(t);
""" % (n, min, max)
        if verbose:
            print(code)
        magma.eval(code)
        return float(magma.eval('s'))
    raise ValueError('unknown system "%s"' % system)



r"""
Jacobians of genus-2 hyperelliptic curves

AUTHORS:

- Sabrina Kunzweiler, Gareth Ma, Giacomo Pope (2024): adapt to smooth model
"""


# ****************************************************************************
#       Copyright (C) 2025 Sabrina Kunzweiler <sabrina.kunzweiler@math.u-bordeaux.fr>
#                     2025 Gareth Ma <grhkm21@gmail.com>
#                     2025 Giacomo Pope <giacomopope@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.schemes.hyperelliptic_curves import (
    jacobian_g2_homset_inert,
    jacobian_g2_homset_ramified,
    jacobian_g2_homset_split,
)
from sage.schemes.hyperelliptic_curves.jacobian_generic import (
    HyperellipticJacobian_generic,
)
from sage.schemes.hyperelliptic_curves.kummer_surface import KummerSurface


class HyperellipticJacobian_g2_generic(HyperellipticJacobian_generic):
    r"""
    Special class to handle optimisations for jacobian computations
    in genus two
    """

    def _point_homset(self, *args, **kwds):
        r"""
        Create the point Hom-set of the Jacobian of a genus-2 curve.

        TODO: make a constructor for this??

        TESTS::

            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(-x^6 + 15*x^4 - 75*x^2 -56, x^3 + x)
            sage: J = Jacobian(H)(QQ)
            sage: type(J) # indirect doctest
            <class 'sage.schemes.hyperelliptic_curves.jacobian_g2_homset_inert.HyperellipticJacobianHomsetInert_g2_with_category'>
        """
        H = self.curve()
        if H.is_ramified():
            return jacobian_g2_homset_ramified.HyperellipticJacobianHomsetRamified_g2(
                *args, **kwds
            )
        if H.is_split():
            return jacobian_g2_homset_split.HyperellipticJacobianHomsetSplit_g2(
                *args, **kwds
            )
        return jacobian_g2_homset_inert.HyperellipticJacobianHomsetInert_g2(
            *args, **kwds
        )

    def kummer_surface(self):
        r"""
        Construct the Kummer surface from the Jacobian of a genus-2 curve.

        INPUT: ``jacobian`` -- the Jacobian of a genus-2 curve

        EXAMPLES::

            sage: R.<x> = GF(13)[]
            sage: H = HyperellipticCurve(x**5 + x)
            sage: J = Jacobian(H)
            sage: K = KummerSurface(J); K
            Kummer Surface of Jacobian of Hyperelliptic Curve over Finite Field of size 13 defined by y^2 = x^5 + x.
            The defining equation is X0^4 - 4*X0*X1^2*X2 + 2*X0^2*X2^2 + X2^4 - 2*X0^2*X1*X3 - 2*X1*X2^2*X3 + X1^2*X3^2 - 4*X0*X2*X3^2
        """
        return KummerSurface(self)



# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR

cdef extern from "gsl/gsl_statistics_double.h":
  double gsl_stats_mean ( double data[],  size_t stride,  size_t n)
  double gsl_stats_variance ( double data[],  size_t stride,  size_t n)
  double gsl_stats_sd ( double data[],  size_t stride,  size_t n)
  double gsl_stats_variance_with_fixed_mean ( double data[],  size_t stride,  size_t n,  double mean)
  double gsl_stats_sd_with_fixed_mean ( double data[],  size_t stride,  size_t n,  double mean)
  double gsl_stats_absdev ( double data[],  size_t stride,  size_t n)
  double gsl_stats_skew ( double data[],  size_t stride,  size_t n)
  double gsl_stats_kurtosis ( double data[],  size_t stride,  size_t n)
  double gsl_stats_lag1_autocorrelation ( double data[],  size_t stride,  size_t n)

  double gsl_stats_covariance ( double data1[],  size_t stride1, double data2[],  size_t stride2,  size_t n)

  double gsl_stats_variance_m ( double data[],  size_t stride,  size_t n,  double mean)
  double gsl_stats_sd_m ( double data[],  size_t stride,  size_t n,  double mean)
  double gsl_stats_absdev_m ( double data[],  size_t stride,  size_t n,  double mean)
  double gsl_stats_skew_m_sd ( double data[],  size_t stride,  size_t n,  double mean,  double sd)
  double gsl_stats_kurtosis_m_sd ( double data[],  size_t stride,  size_t n,  double mean,  double sd)
  double gsl_stats_lag1_autocorrelation_m ( double data[],  size_t stride,  size_t n,  double mean)

  double gsl_stats_covariance_m ( double data1[],  size_t stride1, double data2[],  size_t stride2,  size_t n,  double mean1,  double mean2)

  # DEFINED FOR FLOATING POINT TYPES ONLY

  double gsl_stats_wmean ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n)
  double gsl_stats_wvariance ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n)
  double gsl_stats_wsd ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n)
  double gsl_stats_wvariance_with_fixed_mean ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n,  double mean)
  double gsl_stats_wsd_with_fixed_mean ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n,  double mean)
  double gsl_stats_wabsdev ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n)
  double gsl_stats_wskew ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n)
  double gsl_stats_wkurtosis ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n)

  double gsl_stats_wvariance_m ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n,  double wmean)
  double gsl_stats_wsd_m ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n,  double wmean)
  double gsl_stats_wabsdev_m ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n,  double wmean)
  double gsl_stats_wskew_m_sd ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n,  double wmean,  double wsd)
  double gsl_stats_wkurtosis_m_sd ( double w[],  size_t wstride,  double data[],  size_t stride,  size_t n,  double wmean,  double wsd)


  double gsl_stats_pvariance ( double data1[],  size_t stride1,  size_t n1,  double data2[],  size_t stride2,  size_t n2)
  double gsl_stats_ttest ( double data1[],  size_t stride1,  size_t n1,  double data2[],  size_t stride2,  size_t n2)

  double gsl_stats_max ( double data[],  size_t stride,  size_t n)
  double gsl_stats_min ( double data[],  size_t stride,  size_t n)
  void gsl_stats_minmax (double * min, double * max,  double data[],  size_t stride,  size_t n)

  size_t gsl_stats_max_index ( double data[],  size_t stride,  size_t n)
  size_t gsl_stats_min_index ( double data[],  size_t stride,  size_t n)
  void gsl_stats_minmax_index (size_t * min_index, size_t * max_index,  double data[],  size_t stride,  size_t n)

  double gsl_stats_median_from_sorted_data ( double sorted_data[],  size_t stride,  size_t n)
  double gsl_stats_quantile_from_sorted_data ( double sorted_data[],  size_t stride,  size_t n,  double f)




# sage.doctest: needs sage.groups
r"""
Spinor genus computations.

This file defines the group of spinor operators used
for the computation of spinor genera.
It is meant for internal use only.

EXAMPLES::

    sage: from sage.quadratic_forms.genera.spinor_genus import SpinorOperators
    sage: A = SpinorOperators((2, 3, 7))

AUTHORS:

- Simon Brandhorst (2020-11-1): initial version
"""

# ****************************************************************************
#       Copyright (C) 2020 Simon Brandhorst <sbrandhorst@web.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.groups.abelian_gps.abelian_group_gap import (AbelianGroupGap,
                                                       AbelianGroupElement_gap)
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ


class SpinorOperator(AbelianGroupElement_gap):
    r"""
    A spinor operator seen as a tuple of square classes.

    For `2` the square class is represented as one of `1,3,5,7` and for
    `p` odd it is `1` for a `p`-adic unit square and `-1` for a non-square.

    EXAMPLES::

        sage: from sage.quadratic_forms.genera.spinor_genus import *
        sage: A = SpinorOperators((2, 3, 7))
        sage: A.an_element()
        [2:7, 3:-1, 7:-1]
    """

    def _repr_(self) -> str:
        r"""
        Return the print representation.

        EXAMPLES::

            sage: from sage.quadratic_forms.genera.spinor_genus import *
            sage: A = SpinorOperators((2, 3, 7))
            sage: A.an_element()
            [2:7, 3:-1, 7:-1]
        """
        e = self.exponents()
        p = self.parent()._primes
        s = "[2:"
        if e[0] == 0 == e[1]:
            s += "1"
        elif e[0] == 1 and e[1] == 0:
            s += "3"
        elif e[0] == 0 and e[1] == 1:
            s += "5"
        elif e[0] == 1 == e[1]:
            s += "7"
        for k in range(1, len(p)):
            s += f", {p[k]}:{(-1)**e[k + 1]}"
        s += "]"
        return s


class SpinorOperators(AbelianGroupGap):
    r"""
    The group of spinor operators of a genus.

    It is a product of `p`-adic unit square classes
    used for spinor genus computations.

    INPUT:

    - a tuple of primes `(p_1=2,\dots, p_n`)

    EXAMPLES::

        sage: from sage.quadratic_forms.genera.spinor_genus import *
        sage: SpinorOperators((2, 3, 7))
        Group of SpinorOperators at primes (2, 3, 7)
    """
    def __init__(self, primes):
        r"""
        Initialize the group of spinor operators.

        TESTS::

            sage: from sage.quadratic_forms.genera.spinor_genus import *
            sage: S = SpinorOperators((2, 3, 7))
            sage: TestSuite(S).run()
        """
        if primes[0] != 2:
            raise ValueError("first prime must be 2")
        self._primes = tuple(ZZ(p) for p in primes)
        orders = len(self._primes) * [2] + [2]
        # 3, 5, unit_p1, unit_p2,...
        orders = tuple(orders)
        AbelianGroupGap.__init__(self, orders)

    def __reduce__(self):
        r"""
        Implement pickling.

        OUTPUT:

        a tuple ``f`` such that this element is ``f[0](*f[1])``

        EXAMPLES::

            sage: from sage.quadratic_forms.genera.spinor_genus import SpinorOperators
            sage: S = SpinorOperators((2, 3, 7))
            sage: S == loads(dumps(S))
            True
        """
        return SpinorOperators, (self._primes,)

    Element = SpinorOperator

    def _repr_(self) -> str:
        r"""
        Return the print representation of ``self``.

        EXAMPLES::

            sage: from sage.quadratic_forms.genera.spinor_genus import SpinorOperators
            sage: SpinorOperators((2, 3, 7))
            Group of SpinorOperators at primes (2, 3, 7)
        """
        return f"Group of SpinorOperators at primes {self._primes}"

    def to_square_class(self, x, p):
        r"""
        Return `(1,...,1,x,1,...,1)` with the square class of `x` at position `p`.

        INPUT:

        - ``p`` -- a prime

        - ``x`` -- nonzero rational number

        EXAMPLES::

            sage: from sage.quadratic_forms.genera.spinor_genus import SpinorOperators
            sage: AS = SpinorOperators((2, 3, 7))
            sage: AS.to_square_class(5, 7)
            [2:1, 3:1, 7:-1]
            sage: AS.to_square_class(5, 2)
            [2:5, 3:1, 7:1]
            sage: AS.to_square_class(-5, 2)
            [2:3, 3:1, 7:1]
            sage: AS.to_square_class(7, 2)
            [2:7, 3:1, 7:1]
        """
        x = QQ(x)
        if x == 0:
            raise ValueError("x must be nonzero")
        if p not in self._primes:
            raise ValueError("not a coordinate prime")
        v, u = x.val_unit(p)
        v = v % 2
        if v != 0:
            raise ValueError(f"x(={x}) must be a p-adic unit")
        y = self.one()
        if p == 2:
            u = u % 8
            if u == 3:
                y *= self.gens()[0]
            if u == 5:
                y *= self.gens()[1]
            if u == 7:
                y *= self.gens()[0] * self.gens()[1]
            return y
        i = 1 + self._primes.index(p)
        if not u.is_padic_square(p):
            y *= self.gens()[i]
        return y

    def delta(self, r, prime=None):
        r"""
        Diagonal embedding of rational square classes.

        INPUT:

        - ``r`` -- a nonzero integer; if ``prime`` is ``None``, ``r`` must not
          be divisible by the defining primes of ``self``

        - ``prime`` -- (default: ``None``) a prime or `-1`

        OUTPUT:

        If a prime `p` is given, the method returns
        `\Delta_p(r)`
        otherwise returns `\Delta(r)`
        where both are as defined by Conway-Sloane in
        Chapter 15 9.3 of [CS1988]_.

        EXAMPLES::

            sage: from sage.quadratic_forms.genera.spinor_genus import SpinorOperators
            sage: AS = SpinorOperators((2, 3, 7))
            sage: AS.delta(5)
            [2:5, 3:-1, 7:-1]
            sage: AS.delta(2, prime=3)
            [2:1, 3:-1, 7:1]
            sage: AS.delta(11)
            [2:3, 3:-1, 7:1]
            sage: AS.delta(3, prime=7)
            [2:1, 3:1, 7:-1]
        """
        r = ZZ(r)
        if prime is None:
            if any(p.divides(r) for p in self._primes):
                raise ValueError(f"r must not be divisible by {self._primes}")
            return self.prod([self.to_square_class(r, p)
                              for p in self._primes])
        prime = ZZ(prime)
        if prime == -1:
            r = r.sign()
            return self.prod([self.to_square_class(r, p)
                              for p in self._primes])
        if prime not in self._primes:
            raise ValueError("prime must be among %s" % self._primes)
        v, u = r.val_unit(prime)
        pv = prime**v
        y = self.prod([self.to_square_class(pv, q)
                       for q in self._primes if q != prime])
        if prime in self._primes:
            y *= self.to_square_class(u, p=prime)
        return y



# sage.doctest: needs sage.combinat sage.modules
r"""
`\nu`-Dyck words

A class of the `\nu`-Dyck word, see [PRV2017]_ for details.

This file is based off the class
:func:`DyckWords<sage.combinat.dyck_word.DyckWord>` written by Mike Hansen, Dan
Drake, Florent Hivert, Christian Stump, Mike Zabrocki, Jean--Baptiste Priez
and Travis Scrimshaw.

AUTHORS:

- Aram Dermenjian (2020-09-26)
"""
# ****************************************************************************
#       Copyright (C) 2020 Aram Dermenjian <aram.dermenjian@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from __future__ import annotations
from sage.structure.element import Element
from sage.rings.integer import Integer
from sage.combinat.combinat import CombinatorialElement
from sage.sets.recursively_enumerated_set import RecursivelyEnumeratedSet
from sage.structure.global_options import GlobalOptions
from sage.structure.parent import Parent
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.misc.latex import latex

from sage.combinat.words.paths import WordPaths_north_east
from sage.combinat.words.paths import FiniteWordPath_north_east

ndw_open_symbol = 1
ndw_close_symbol = 0


def update_ndw_symbols(os, cs):
    r"""
    A way to alter the open and close symbols from sage.

    INPUT:

    - ``os`` -- the open symbol
    - ``cs`` -- the close symbol

    EXAMPLES::

        sage: from sage.combinat.nu_dyck_word import update_ndw_symbols
        sage: update_ndw_symbols(0,1)
        sage: dw = NuDyckWord('0101001','0110010'); dw
        [0, 1, 0, 1, 0, 0, 1]

        sage: dw = NuDyckWord('1010110','1001101'); dw
        Traceback (most recent call last):
        ...
        ValueError: invalid nu-Dyck word
        sage: update_ndw_symbols(1,0)
    """
    global ndw_open_symbol
    global ndw_close_symbol
    ndw_open_symbol = os
    ndw_close_symbol = cs


def replace_dyck_char(x):
    r"""
    A map sending an opening character (``'1'``, ``'N'``, and ``'('``) to
    ``ndw_open_symbol``, a closing character (``'0'``, ``'E'``, and ``')'``) to
    ``ndw_close_symbol``, and raising an error on any input other than one of
    the opening or closing characters.

    This is the inverse map of :func:`replace_dyck_symbol`.

    INPUT:

    - ``x`` -- string; a ``'1'``, ``'0'``, ``'N'``, ``'E'``, ``'('`` or ``')'``

    OUTPUT:

    - If ``x`` is an opening character, replace ``x`` with the
      constant ``ndw_open_symbol``.

    - If ``x`` is a closing character, replace ``x`` with the
      constant ``ndw_close_symbol``.

    - Raise a :exc:`ValueError` if ``x`` is neither an opening nor a
      closing character.

    .. SEEALSO:: :func:`replace_dyck_symbol`

    EXAMPLES::

        sage: from sage.combinat.nu_dyck_word import replace_dyck_char
        sage: replace_dyck_char('(')
        1
        sage: replace_dyck_char(')')
        0
        sage: replace_dyck_char(1)
        Traceback (most recent call last):
        ...
        ValueError
    """
    if x == '(' or x == 'N' or x == str(ndw_open_symbol):
        return ndw_open_symbol
    if x == ')' or x == 'E' or x == str(ndw_close_symbol):
        return ndw_close_symbol
    raise ValueError


def replace_dyck_symbol(x, open_char='N', close_char='E') -> str:
    r"""
    A map sending ``ndw_open_symbol`` to ``open_char`` and ``ndw_close_symbol``
    to ``close_char``, and raising an error on any input other than
    ``ndw_open_symbol`` and ``ndw_close_symbol``. The values of the constants
    ``ndw_open_symbol`` and ``ndw_close_symbol`` are subject to change.

    This is the inverse map of :func:`replace_dyck_char`.

    INPUT:

    - ``x`` -- either ``ndw_open_symbol`` or ``ndw_close_symbol``

    - ``open_char`` -- string (optional) default ``'N'``

    - ``close_char`` -- string (optional) default ``'E'``

    OUTPUT: if ``x`` is ``ndw_open_symbol``, replace ``x`` with ``open_char``

    - If ``x`` is ``ndw_close_symbol``, replace ``x`` with ``close_char``.

    - If ``x`` is neither ``ndw_open_symbol`` nor ``ndw_close_symbol``, a
      :exc:`ValueError` is raised.

    .. SEEALSO:: :func:`replace_dyck_char`

    EXAMPLES::

        sage: from sage.combinat.nu_dyck_word import replace_dyck_symbol
        sage: replace_dyck_symbol(1)
        'N'
        sage: replace_dyck_symbol(0)
        'E'
        sage: replace_dyck_symbol(3)
        Traceback (most recent call last):
        ...
        ValueError
    """
    if x == ndw_open_symbol:
        return open_char
    if x == ndw_close_symbol:
        return close_char
    raise ValueError


class NuDyckWord(CombinatorialElement):
    r"""
    A `\nu`-Dyck word.

    Given a lattice path `\nu` in the `\ZZ^2` grid starting at the origin
    `(0,0)` consisting of North `N = (0,1)` and East `E = (1,0)` steps, a
    `\nu`-Dyck path is a lattice path in the `\ZZ^2` grid starting at the
    origin `(0,0)` and ending at the same coordinate as `\nu` such that it is
    weakly above `\nu`. A `\nu`-Dyck word is the representation of a
    `\nu`-Dyck path where a North step is represented by a 1 and an East step
    is represented by a 0.

    INPUT:

    - ``k1`` -- a path for the `\nu`-Dyck word

    - ``k2`` -- a path for `\nu`

    EXAMPLES::

        sage: dw = NuDyckWord([1,0,1,0],[1,0,0,1]); dw
        [1, 0, 1, 0]
        sage: print(dw)
        NENE
        sage: dw.height()
        2

        sage: dw = NuDyckWord('1010',[1,0,0,1]); dw
        [1, 0, 1, 0]

        sage: dw = NuDyckWord('NENE',[1,0,0,1]); dw
        [1, 0, 1, 0]

        sage: NuDyckWord([1,0,1,0],[1,0,0,1]).pretty_print()
           __
         _|x
        | . .

        sage: from sage.combinat.nu_dyck_word import update_ndw_symbols
        sage: update_ndw_symbols(0,1)
        sage: dw = NuDyckWord('0101001','0110010'); dw
        [0, 1, 0, 1, 0, 0, 1]
        sage: dw.pp()
             __
            |x
           _| .
         _|x  .
        | . . .
        sage: update_ndw_symbols(1,0)
    """
    @staticmethod
    def __classcall_private__(cls, dw=None, nu=None, **kwargs):
        """
        Return an element with the appropriate parent.

        EXAMPLES::

            sage: NuDyckWord('110100','101010')
            [1, 1, 0, 1, 0, 0]
            sage: NuDyckWord('010','010')
            [0, 1, 0]
        """
        # if dw is none, then we might have a normal Dyck word
        if dw is None:
            from sage.combinat.dyck_word import DyckWord
            return DyckWord(dw, kwargs)

        if isinstance(dw, NuDyckWord):
            return dw

        if nu is None:
            raise ValueError("nu required")

        dw = to_word_path(dw)
        nu = to_word_path(nu)

        if path_weakly_above_other(dw, nu):
            return NuDyckWords(nu)(dw)

        raise ValueError("invalid nu-Dyck word")

    def __init__(self, parent, dw, latex_options=None):
        """
        Initialize a nu-Dyck word.

        EXAMPLES::

            sage: NuDyckWord('010', '010')
            [0, 1, 0]
            sage: NuDyckWord('110100','101010')
            [1, 1, 0, 1, 0, 0]
        """
        Element.__init__(self, parent)
        self._path = to_word_path(dw)
        self._list = list(self._path)

        if parent is None:
            raise ValueError("need parent object")

        self._nu = parent._nu

        if latex_options is None:
            latex_options = {}
        self._latex_options = dict(latex_options)

    def __eq__(self, other):
        """
        Return if two paths are equal.

        EXAMPLES::

            sage: u = NuDyckWord('010','010')
            sage: w = NuDyckWord('110100','101010')
            sage: w == w
            True
            sage: u == w
            False
            sage: u == 4
            False
        """
        if not isinstance(other, NuDyckWord):
            return False
        return self._path == other._path and self._nu == other._nu

    def __neq__(self, other):
        """
        Return if two paths are not equal.

        EXAMPLES::

            sage: u = NuDyckWord('010','010')
            sage: w = NuDyckWord('110100','101010')
            sage: w != w
            False
            sage: u != w
            True
            sage: u != 4
            True
        """
        return not self.__eq__(other)

    def __le__(self, other):
        """
        Return if one path is included in another.

        EXAMPLES::

            sage: ND1 = NuDyckWord('101', '011')
            sage: ND2 = NuDyckWord('110', '011')
            sage: ND3 = NuDyckWord('011', '011')
            sage: ND1 <= ND1
            True
            sage: ND1 <= ND2
            True
            sage: ND2 <= ND1
            False
            sage: ND3 <= ND2
            True
            sage: ND3 <= ND1
            True
        """
        if self._nu == other._nu:
            return path_weakly_above_other(other._path, self._path)
        return False

    def __lt__(self, other):
        """
        Return if one path is strictly included in another.

        EXAMPLES::

            sage: ND1 = NuDyckWord('101', '011')
            sage: ND2 = NuDyckWord('110', '011')
            sage: ND3 = NuDyckWord('011', '011')
            sage: ND1 < ND1
            False
            sage: ND1 < ND2
            True
            sage: ND2 < ND1
            False
            sage: ND3 < ND2
            True
            sage: ND3 < ND1
            True
        """
        return self.__le__(other) and not self.__eq__(other)

    def __ge__(self, other):
        """
        Return if one path is included in another.

        EXAMPLES::

            sage: ND1 = NuDyckWord('101', '011')
            sage: ND2 = NuDyckWord('110', '011')
            sage: ND3 = NuDyckWord('011', '011')
            sage: ND1 >= ND1
            True
            sage: ND1 >= ND2
            False
            sage: ND2 >= ND1
            True
            sage: ND1 >= ND3
            True
            sage: ND2 >= ND3
            True
        """
        if self._nu == other._nu:
            return path_weakly_above_other(self._path, other._path)
        return False

    def __gt__(self, other):
        """
        Return if one path is strictly included in another.

        EXAMPLES::

            sage: ND1 = NuDyckWord('101', '011')
            sage: ND2 = NuDyckWord('110', '011')
            sage: ND3 = NuDyckWord('011', '011')
            sage: ND1 > ND1
            False
            sage: ND1 > ND2
            False
            sage: ND2 > ND1
            True
            sage: ND1 > ND3
            True
            sage: ND2 > ND3
            True
        """
        return self.__ge__(other) and not self.__eq__(other)

    def _cache_key(self) -> tuple:
        """
        Return a cache key for ``self``.

        EXAMPLES::

            sage: u = NuDyckWord('010','010')
            sage: u._cache_key()
            (0, 1, 0, 0, 1, 0)
        """
        return tuple(self._path) + tuple(self._nu)

    def __hash__(self) -> int:
        """
        Return a hash for ``self``.

        EXAMPLES::

            sage: u = NuDyckWord('010','010')
            sage: hash(u)  # random
            -4577085166836515071
        """
        return hash(''.join(str(i) for i in self._list))

    def set_latex_options(self, D):
        r"""
        Set the latex options for use in the ``_latex_`` function.

        The default values are set in the ``__init__`` function.

        - ``color`` -- (default: black) the line color

        - ``line width`` -- (default: `2 \times` ``tikz_scale``) value
          representing the line width

        - ``nu_options`` -- (default: ``'rounded corners=1, color=red, line
          width=1'``) string to indicate what the tikz options should be for
          path of `\nu`

        - ``points_color`` -- (default: ``'black'``) str to indicate color
          points should be drawn with

        - ``show_grid`` -- boolean (default: ``True``); value to indicate if
          grid should be shown

        - ``show_nu`` -- boolean (default: ``True``); value to indicate if `\nu`
          should be shown

        - ``show_points`` -- boolean (default: ``False``); value to indicate
          if points should be shown on path

        - ``tikz_scale`` -- (default: 1) scale for use with the tikz package

        INPUT:

        - ``D`` -- dictionary with a list of latex parameters to change

        EXAMPLES::

            sage: NDW = NuDyckWord('010','010')
            sage: NDW.set_latex_options({"tikz_scale":2})
            sage: NDW.set_latex_options({"color":"blue", "show_points":True})

        .. TODO::

            This should probably be merged into NuDyckWord.options.
        """
        for opt in D:
            self._latex_options[opt] = D[opt]

    def latex_options(self) -> dict:
        r"""
        Return the latex options for use in the ``_latex_`` function as a
        dictionary.

        The default values are set using the options.

        - ``color`` -- (default: black) the line color

        - ``line width`` -- (default: 2*``tikz_scale``) value representing the
          line width

        - ``nu_options`` -- (default: ``'rounded corners=1, color=red, line
          width=1'``) string to indicate what the tikz options should be for
          path of `\nu`

        - ``points_color`` -- (default: ``'black'``) str to indicate color
          points should be drawn with

        - ``show_grid`` -- boolean (default: ``True``); value to indicate if
          grid should be shown

        - ``show_nu`` -- boolean (default: ``True``); value to indicate if `\nu`
          should be shown

        - ``show_points`` -- boolean (default: ``False``); value to indicate
          if points should be shown on path

        - ``tikz_scale`` -- (default: 1) scale for use with the tikz package

        EXAMPLES::

            sage: NDW = NuDyckWord('010','010')
            sage: NDW.latex_options()
            {'color': black,
             'line width': 2,
             'nu_options': rounded corners=1, color=red, line width=1,
             'points_color': black,
             'show_grid': True,
             'show_nu': True,
             'show_points': False,
             'tikz_scale': 1}

        .. TODO::

            This should probably be merged into NuDyckWord.options.
        """
        d = self._latex_options.copy()
        opts = self.parent().options
        if "tikz_scale" not in d:
            d["tikz_scale"] = opts.latex_tikz_scale
        if "line width" not in d:
            d["line width"] = opts.latex_line_width_scalar * d["tikz_scale"]
        if "color" not in d:
            d["color"] = opts.latex_color
        if "show_points" not in d:
            d["show_points"] = opts.latex_show_points
        if "points_color" not in d:
            d["points_color"] = opts.latex_points_color
        if "show_grid" not in d:
            d["show_grid"] = opts.latex_show_grid
        if "show_nu" not in d:
            d["show_nu"] = opts.latex_show_nu
        if "nu_options" not in d:
            d["nu_options"] = opts.latex_nu_options
        return d

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self`` depending on
        :meth:`NuDyckWords.options`.

        TESTS::

            sage: NuDyckWord('01010','00011')
            [0, 1, 0, 1, 0]
            sage: NuDyckWord('10010','00011')
            [1, 0, 0, 1, 0]
            sage: NuDyckWords.options.display="lattice"
            sage: NuDyckWord('10010','00011')
                 __
             ___|x
            |x x x

            sage: NuDyckWords.options._reset()
        """
        return self.parent().options._dispatch(self, '_repr_', 'display')

    def _repr_list(self) -> str:
        r"""
        Return a string representation of ``self`` as a list.

        TESTS::

            sage: NuDyckWord([1,1,0],[1,0,1])  # indirect doctest
            [1, 1, 0]
            sage: NuDyckWord('NNEE','NENE')
            [1, 1, 0, 0]
        """
        return str(list(self._path))

    def _repr_lattice(self, style=None, labelling=None):
        r"""
        See :meth:`pretty_print()`.

        TESTS::

            sage: n = NuDyckWord('00011001000100','00011001000100')
            sage: print(n._repr_lattice(style='N-E', labelling=[1,2,3,4]))
                             ____
                       _____| . . 4
                   ___| . . . . . 3
                  | . . . . . . . 2
            ______| . . . . . . . 1


            sage: print(NuDyckWord('100101','010011')._repr_lattice())
                 _|
             ___|x
            |x  . .

            sage: print(NuDyckWord('110010','001011')._repr_lattice())
                 __
             ___|x
            |x x x
            |x x  .
        """
        if style is None:
            style = self.parent().options.diagram_style
            if style == "grid":
                style = "N-E"

        if style == "N-E":
            path_length = self._path.length()
            height = self._path.height()
            width = self._path.width()
            if path_length == 0:
                return ".\n"

            # Handle right-hand side labels
            if labelling is None:
                labels = [" "] * height
            else:
                if len(labelling) != height:
                    raise ValueError(f"the given labelling has the wrong length: {height} needed")
                labels = [str(label) for label in labelling]
                max_length = max(len(label) for label in labels)
                labels = [lbl.rjust(max_length + 1) for lbl in labels]

            rev_path = list(self._path.reversal())
            rev_nu_path = list(self._nu.reversal())
            ts = ""

            # Grab first line
            cur_pos = rev_path.index(ndw_open_symbol)
            cur_nu_pos = rev_nu_path.index(ndw_open_symbol)
            if cur_pos > 0:
                ts += "  " * (width - cur_pos)
                ts += " _" + "__" * (cur_pos - 1)
                ts += "_\n"

            # Middle Lines
            for i in range(height - 1):
                old_pos = cur_pos
                old_nu_pos = cur_nu_pos
                cur_pos = rev_path.index(ndw_open_symbol, cur_pos + 1)
                cur_nu_pos = rev_nu_path.index(ndw_open_symbol, cur_nu_pos + 1)

                ts += "  " * (width - cur_pos + i + 1)
                if cur_pos != old_pos + 1:
                    ts += " _" + "__" * (cur_pos - old_pos - 2)
                ts += "|"
                if old_pos >= 0:
                    ts += "x " * (old_pos - old_nu_pos)
                    ts += " ." * (old_nu_pos - i)
                ts += labels[height - i - 1]
                ts += "\n"

            # Final line
            ts += "__" * (path_length - cur_pos - 1)
            ts += "|"
            ts += "x " * (cur_pos - cur_nu_pos)
            ts += " ." * (cur_nu_pos - i - 1)
            ts += labels[0]
            ts += "\n"
            return ts
        raise ValueError(f"the given style (={style}) is not valid")

    def _ascii_art_(self):
        r"""
        Return an ASCII art representation of ``self``.

        TESTS::

            sage: ascii_art(NuDyckWord('00011001000100','00011001000100'))
                             ____
                       _____| . .
                   ___| . . . . .
                  | . . . . . . .
            ______| . . . . . . .
        """
        from sage.typeset.ascii_art import AsciiArt
        rep = self.parent().options.ascii_art
        if rep == "pretty_output":
            ret = self._repr_lattice()
        return AsciiArt(ret.splitlines(), baseline=0)

    def __str__(self) -> str:
        r"""
        Return a string consisting of N and E steps corresponding to
        the `\nu`-Dyck word.

        EXAMPLES::

            sage: str(NuDyckWord('100101','010011'))
            'NEENEN'
            sage: str(NuDyckWord('101010','100110'))
            'NENENE'
        """
        return "".join(replace_dyck_symbol(let) for let in self._path)

    def pretty_print(self, style=None, labelling=None):
        r"""
        Display a NuDyckWord as a lattice path in the `\ZZ^2` grid.

        If the ``style`` is "N-E", then a cell below the diagonal is
        indicated by a period, whereas a cell below the path but above
        the diagonal is indicated by an x. If a list of labels is
        included, they are displayed along the vertical edges of the
        Dyck path.

        INPUT:

        - ``style`` -- (default: ``None``) can either be:

          - ``None`` to use the option default
          - "N-E" to show ``self`` as a path of north and east steps, or

        - ``labelling`` -- (if style is "N-E") a list of labels assigned to
          the up steps in ``self``

        - ``underpath`` -- (if style is "N-E", default: ``True``) if ``True``,
          an ``x`` to show the boxes between `\nu` and the `\nu`-Dyck Path

        EXAMPLES::

            sage: for ND in NuDyckWords('101010'): ND.pretty_print()
                 __
               _| .
             _| . .
            | . . .
                 __
             ___| .
            |x  . .
            | . . .
               ____
              |x  .
             _| . .
            | . . .
               ____
             _|x  .
            |x  . .
            | . . .
             ______
            |x x  .
            |x  . .
            | . . .

        ::

            sage: nu = [1,0,1,0,1,0,1,0,1,0,1,0]
            sage: ND = NuDyckWord([1,1,1,0,1,0,0,1,1,0,0,0],nu)
            sage: ND.pretty_print()
                   ______
                  |x x  .
               ___|x  . .
             _|x x  . . .
            |x x  . . . .
            |x  . . . . .
            | . . . . . .

        ::

            sage: NuDyckWord([1,1,0,0,1,0],[1,0,1,0,1,0]).pretty_print(
            ....: labelling=[1,3,2])
                 __
             ___| . 2
            |x  . . 3
            | . . . 1

        ::

            sage: NuDyckWord('1101110011010010001101111000110000',
            ....: '1010101010101010101010101010101010').pretty_print(
            ....: labelling=list(range(1,18)))
                                       ________
                                      |x x x  . 17
                                 _____|x x  . . 16
                                |x x x x  . . . 15
                                |x x x  . . . . 14
                                |x x  . . . . . 13
                               _|x  . . . . . . 12
                              |x  . . . . . . . 11
                         _____| . . . . . . . . 10
                     ___|x x  . . . . . . . . .  9
                   _|x x x  . . . . . . . . . .  8
                  |x x x  . . . . . . . . . . .  7
               ___|x x  . . . . . . . . . . . .  6
              |x x x  . . . . . . . . . . . . .  5
              |x x  . . . . . . . . . . . . . .  4
             _|x  . . . . . . . . . . . . . . .  3
            |x  . . . . . . . . . . . . . . . .  2
            | . . . . . . . . . . . . . . . . .  1


        ::

            sage: NuDyckWord().pretty_print()
            .
        """
        print(self._repr_lattice(style, labelling))

    pp = pretty_print

    def _latex_(self):
        r"""
        A latex representation of ``self`` using the tikzpicture package.

        EXAMPLES::

            sage: NDW = NuDyckWord('010','010')
            sage: NDW.set_latex_options({"show_points":True})
            sage: latex(NDW)
            \vcenter{\hbox{$\begin{tikzpicture}[scale=1]
              \draw[dotted] (0, 0) grid (2, 1);
              \draw[line width=2,color=black,fill=black](0, 0) circle (0.21);
              \draw[line width=2,color=black,fill=black](1, 0) circle (0.21);
              \draw[line width=2,color=black,fill=black](1, 1) circle (0.21);
              \draw[line width=2,color=black,fill=black](2, 1) circle (0.21);
              \draw[rounded corners=1, color=red, line width=1] (0, 0) -- (1, 0) -- (1, 1) -- (2, 1);
              \draw[rounded corners=1, color=black, line width=2] (0, 0) -- (1, 0) -- (1, 1) -- (2, 1);
            \end{tikzpicture}$}}
            sage: NuDyckWord('01','01')._latex_()
            '\\vcenter{\\hbox{$\\begin{tikzpicture}[scale=1]\n  \\draw[dotted] (0, 0) grid (1, 1);\n  \\draw[rounded corners=1, color=red, line width=1] (0, 0) -- (1, 0) -- (1, 1);\n  \\draw[rounded corners=1, color=black, line width=2] (0, 0) -- (1, 0) -- (1, 1);\n\\end{tikzpicture}$}}'
            sage: NuDyckWord('101100','101010')._latex_()
            '\\vcenter{\\hbox{$\\begin{tikzpicture}[scale=1]\n  \\draw[dotted] (0, 0) grid (3, 3);\n  \\draw[rounded corners=1, color=red, line width=1] (0, 0) -- (0, 1) -- (1, 1) -- (1, 2) -- (2, 2) -- (2, 3) -- (3, 3);\n  \\draw[rounded corners=1, color=black, line width=2] (0, 0) -- (0, 1) -- (1, 1) -- (1, 2) -- (1, 3) -- (2, 3) -- (3, 3);\n\\end{tikzpicture}$}}'
        """
        latex.add_package_to_preamble_if_available("tikz")
        latex_options = self.latex_options()

        # Start setting up tikz
        res = "\\vcenter{\\hbox{$\\begin{tikzpicture}"
        res += "[scale=" + str(latex_options['tikz_scale']) + "]"
        res += "\n"

        # Setup background grid
        if latex_options['show_grid']:
            grid = [((0, 0), (self.width(), self.height()))]
            for v1, v2 in grid:
                res += f"  \\draw[dotted] {v1} grid {v2};"
                res += "\n"

        # Add points if wanted
        if latex_options['show_points']:
            pt_color = latex_options['points_color']
            radius = 0.15 + .03 * latex_options['line width']
            for v in self.points():
                res += "  \\draw[line width=2,"
                res += f"color={pt_color},fill={pt_color}]"
                res += f"{v} circle ({radius});"
                res += "\n"

        # Add nu if wanted
        if latex_options['show_nu']:
            res += "  \\draw[%s]" % (str(latex_options['nu_options']))
            for k, p in enumerate(self._nu.points()):
                if k == 0:
                    res += " %s" % (str(p))
                else:
                    res += " -- %s" % (str(p))
            res += ";\n"

        # setup Path
        res += "  \\draw[rounded corners=1, color={}, line width={}]".format(
            latex_options['color'],
            str(latex_options['line width'])
        )
        for k, p in enumerate(self._path.points()):
            if k == 0:
                res += " %s" % (str(p))
            else:
                res += " -- %s" % (str(p))
        res += ";\n"
        res += "\\end{tikzpicture}$}}"
        return res

    def plot(self, **kwds):
        r"""
        Plot a `\nu`-Dyck word as a continuous path.

        EXAMPLES::

            sage: NDW = NuDyckWord('010','010')
            sage: NDW.plot()                                                            # needs sage.plot
            Graphics object consisting of 1 graphics primitive
        """
        from sage.plot.plot import list_plot
        return list_plot(list(self.points()), plotjoined=True, **kwds)

    def path(self):
        r"""
        Return the underlying path object.

        EXAMPLES::

            sage: NDW = NuDyckWord('10011001000','00100101001')
            sage: NDW.path()
            Path: 10011001000
        """
        return self._path

    def height(self):
        r"""
        Return the height of ``self``.

        The height is the number of ``north`` steps.

        EXAMPLES::

            sage: NuDyckWord('1101110011010010001101111000110000',
            ....: '1010101010101010101010101010101010').height()
            17
        """
        return self._path.height()

    def width(self):
        r"""
        Return the width of ``self``.

        The width is the number of ``east`` steps.

        EXAMPLES::

            sage: NuDyckWord('110111001101001000110111100011000',
            ....: '101010101010101010101010101010101').width()
            16
        """
        return self._path.width()

    def length(self):
        r"""
        Return the length of ``self``.

        The length is the total number of steps.

        EXAMPLES::

            sage: NDW = NuDyckWord('10011001000','00100101001')
            sage: NDW.length()
            11
        """
        return self._path.length()

    def points(self):
        r"""
        Return an iterator for the points on the `\nu`-Dyck path.

        EXAMPLES::

            sage: list(NuDyckWord('110111001101001000110111100011000',
            ....: '101010101010101010101010101010101')._path.points())
            [(0, 0),
             (0, 1),
             (0, 2),
             (1, 2),
             (1, 3),
             (1, 4),
             (1, 5),
             (2, 5),
             (3, 5),
             (3, 6),
             (3, 7),
             (4, 7),
             (4, 8),
             (5, 8),
             (6, 8),
             (6, 9),
             (7, 9),
             (8, 9),
             (9, 9),
             (9, 10),
             (9, 11),
             (10, 11),
             (10, 12),
             (10, 13),
             (10, 14),
             (10, 15),
             (11, 15),
             (12, 15),
             (13, 15),
             (13, 16),
             (13, 17),
             (14, 17),
             (15, 17),
             (16, 17)]
        """
        return self._path.points()

    def heights(self):
        r"""
        Return the heights of each point on ``self``.

        We view the Dyck word as a Dyck path from `(0,0)` to
        `(x,y)` in the first quadrant by letting ``1``'s represent
        steps in the direction `(0,1)` and ``0``'s represent steps in
        the direction `(1,0)`.

        The heights is the sequence of the `y`-coordinates of all
        `x+y` lattice points along the path.

        EXAMPLES::

            sage: NuDyckWord('010','010').heights()
            [0, 0, 1, 1]
            sage: NuDyckWord('110100','101010').heights()
            [0, 1, 2, 2, 3, 3, 3]
        """
        return self._path.height_vector()

    def widths(self):
        r"""
        Return the widths of each point on ``self``.

        We view the Dyck word as a Dyck path from `(0,0)` to
        `(x,y)` in the first quadrant by letting ``1``'s represent
        steps in the direction `(0,1)` and ``0``'s represent steps in
        the direction `(1,0)`.

        The widths is the sequence of the `x`-coordinates of all
        `x+y` lattice points along the path.

        EXAMPLES::

            sage: NuDyckWord('010','010').widths()
            [0, 1, 1, 2]
            sage: NuDyckWord('110100','101010').widths()
            [0, 0, 0, 1, 1, 2, 3]
        """
        return self._path.width_vector()

    def horizontal_distance(self):
        r"""
        Return a list of how far each point is from `\nu`.

        EXAMPLES::

            sage: NDW = NuDyckWord('10010100','00000111')
            sage: NDW.horizontal_distance()
            [5, 5, 4, 3, 3, 2, 2, 1, 0]
            sage: NDW = NuDyckWord('10010100','00001011')
            sage: NDW.horizontal_distance()
            [4, 5, 4, 3, 3, 2, 2, 1, 0]
            sage: NDW = NuDyckWord('10011001000','00100101001')
            sage: NDW.horizontal_distance()
            [2, 4, 3, 2, 3, 5, 4, 3, 3, 2, 1, 0]
        """
        # Grab furthest east point at each height of nu
        nu_points = list(self._nu.points())
        nu_easts = [max(i for i, j in nu_points if j == k)
                    for k in range(self._nu.height() + 1)]

        points = list(self._path.points())
        return [nu_easts[j] - i for i, j in points]

    def can_mutate(self, i) -> bool | int:
        """
        Return True/False based off if mutable at height `i`.

        Can only mutate if an east step is followed by a north step at height
        `i`.

        OUTPUT: whether we can mutate at height of `i`

        EXAMPLES::

            sage: NDW = NuDyckWord('10010100','00000111')
            sage: NDW.can_mutate(1)
            False
            sage: NDW.can_mutate(3)
            5

        TESTS::

            sage: NDW = NuDyckWord('10010100','00000111')
            sage: NDW.can_mutate(33)
            Traceback (most recent call last):
            ...
            ValueError: cannot mutate above or below path
        """
        if i > self.height() or i <= 0:
            raise ValueError('cannot mutate above or below path')

        # Find the ith north step
        level = 0
        ndw = self._list
        for j, k in enumerate(ndw):
            if k == ndw_open_symbol:
                level += 1
                if level == i:
                    break
        if j > 0 and ndw[j - 1] == ndw_close_symbol:
            return j
        return False

    def mutate(self, i) -> NuDyckWord | None:
        r"""
        Return a new `\nu`-Dyck Word if possible.

        If at height `i` we have an east step E meeting a north step N then we
        calculate all horizontal distances from this point until we find
        the first point that has the same horizontal distance to `\nu`. We let

        - d is everything up until EN (not including EN)

        - f be everything between N and the point with the same horizontal
          distance (including N)

        - g is everything after f

        .. SEEALSO:: :meth:`can_mutate`

        EXAMPLES::

            sage: NDW = NuDyckWord('10010100','00000111')
            sage: NDW.mutate(1)
            sage: NDW.mutate(3)
            [1, 0, 0, 1, 1, 0, 0, 0]
        """
        mutation_index = self.can_mutate(i)
        if not mutation_index:
            return None

        horiz = self.horizontal_distance()
        # Find horiz in between East and North Step
        horiz_num = horiz[mutation_index]
        other_index = len(horiz)
        for i in range(mutation_index + 1, len(horiz)):
            if horiz[i] == horiz_num:
                other_index = i
                break
        ndw = self._list
        d = ndw[0:mutation_index - 1]
        e = ndw[mutation_index:other_index]
        f = ndw[other_index:]
        return NuDyckWord(d + e + [ndw_close_symbol] + f, self._nu)


class NuDyckWords(Parent):
    r"""
    `\nu`-Dyck words.

    Given a lattice path `\nu` in the `\ZZ^2` grid starting at the origin
    `(0,0)` consisting of North `N = (0,1)` and East `E = (1,0)` steps, a
    `\nu`-Dyck path is a lattice path in the`\ZZ^2` grid starting at the
    origin `(0,0)` and ending at the same coordinate as `\nu` such that it is
    weakly above `\nu`. A `\nu`-Dyck word is the representation of a
    `\nu`-Dyck path where a North step is represented by a 1 and an East step
    is represented by a 0.

    INPUT:

    - ``nu`` -- the base lattice path

    EXAMPLES::

        sage: NDW = NuDyckWords('1010'); NDW
        [1, 0, 1, 0] Dyck words
        sage: [1,0,1,0] in NDW
        True
        sage: [1,1,0,0] in NDW
        True
        sage: [1,0,0,1] in NDW
        False
        sage: [0,1,0,1] in NDW
        False
        sage: NDW.cardinality()
        2
    """

    Element = NuDyckWord

    def __init__(self, nu=()) -> None:
        """
        Initialize ``self``.

        EXAMPLES::

            sage: TestSuite(NuDyckWords(nu=[1,0,1])).run()
        """
        Parent.__init__(self, category=FiniteEnumeratedSets())

        self._nu = to_word_path(nu)
        if self._nu is None:
            raise ValueError("invalid nu supplied")

    # add options to class
    class options(GlobalOptions):
        r"""
        Set and display the options for `\nu`-Dyck words. If no parameters
        are set, then the function returns a copy of the options dictionary.

        The ``options`` to `\nu`-Dyck words can be accessed as the method
        :meth:`NuDyckWords.options` of :class:`NuDyckWords` and
        related parent classes.

        @OPTIONS

        EXAMPLES::

            sage: ND = NuDyckWords('101')
            sage: ND
            [1, 0, 1] Dyck words
            sage: ND.options
            Current options for NuDyckWords
              - ascii_art:               pretty_output
              - diagram_style:           grid
              - display:                 list
              - latex_color:             black
              - latex_line_width_scalar: 2
              - latex_nu_options:        rounded corners=1, color=red, line width=1
              - latex_points_color:      black
              - latex_show_grid:         True
              - latex_show_nu:           True
              - latex_show_points:       False
              - latex_tikz_scale:        1
        """
        NAME = 'NuDyckWords'
        module = 'sage.combinat.nu_dyck_path'
        display = {'default': "list",
                   'description': 'Specifies how nu Dyck words should be printed',
                   'values': {'list': 'displayed as a list',
                              'lattice': 'displayed on the lattice defined by ``diagram_style``'},
                   'case_sensitive': False}
        ascii_art = {'default': "pretty_output",
                     'description': 'Specifies how the ascii art of nu Dyck words should be printed',
                     'values': {'pretty_output': "Using pretty printing"},
                     'alias': {'pretty_print': "pretty_output"},
                     'case_sensitive': False}
        diagram_style = {'default': "grid",
                         'values': {
                             'grid': 'printing as paths on a grid using N and E steps'},
                         'alias': {'N-E': 'grid'},
                         'case_sensitive': False}
        latex_tikz_scale = {'default': 1,
                            'description': 'The default value for the tikz scale when latexed',
                            'checker': lambda x: True}  # More trouble than it's worth to check
        latex_line_width_scalar = {'default': 2,
                                   'description': 'The default value for the line width as a '
                                   'multiple of the tikz scale when latexed',
                                   'checker': lambda x: True}  # More trouble than it's worth to check
        latex_color = {'default': "black",
                       'description': 'The default value for the color when latexed',
                       'checker': lambda x: isinstance(x, str)}
        latex_show_points = {'default': False,
                             'description': 'The default value for showing points',
                             'checker': lambda x: isinstance(x, bool)}
        latex_points_color = {'default': 'black',
                              'description': 'The default value for path color.',
                              'checker': lambda x: isinstance(x, str)}
        latex_show_grid = {'default': True,
                           'description': 'The default value for showing grid',
                           'checker': lambda x: isinstance(x, bool)}
        latex_show_nu = {'default': True,
                         'description': 'The default value for showing nu',
                         'checker': lambda x: isinstance(x, bool)}
        latex_nu_options = {'default': 'rounded corners=1, color=red, line width=1',
                            'description': 'The default value for options for nu path',
                            'checker': lambda x: isinstance(x, str)}

    def _element_constructor_(self, word):
        """
        Construct an element of ``self``.

        EXAMPLES::

            sage: NDW = NuDyckWords('101')
            sage: elt = NDW('110'); elt
            [1, 1, 0]
            sage: elt.parent() is NDW
            True
        """
        if isinstance(word, NuDyckWord) and word.parent() is self:
            return word
        return self.element_class(self, to_word_path(word))

    def __contains__(self, x) -> bool:
        r"""
        Check for containment.

        TESTS::

            sage: NDW = NuDyckWords([1,0,1,1])
            sage: [1,1,0,1] in NDW
            True
            sage: [1,0,1,1] in NDW
            True
            sage: [0] in NDW
            False
            sage: [1, 0] in NDW
            False
        """
        return path_weakly_above_other(to_word_path(x), self._nu)

    def __eq__(self, other):
        """
        Return equality.

        TESTS::

            sage: A = NuDyckWords([1,0,1,1])
            sage: B = NuDyckWords([1,0,1,1])
            sage: C = NuDyckWords([1,0,1,1,1])
            sage: A == B
            True
            sage: A == C
            False
        """
        if not isinstance(other, NuDyckWords):
            return False
        return self._nu == other._nu

    def __neq__(self, other):
        """
        Return inequality.

        TESTS::

            sage: A = NuDyckWords([1,0,1,1])
            sage: B = NuDyckWords([1,0,1,1])
            sage: C = NuDyckWords([1,0,1,1,1])
            sage: A != B
            False
            sage: A != C
            True
        """
        return not self.__eq__(other)

    def _repr_(self) -> str:
        r"""
        TESTS::

            sage: NuDyckWords([1,0,1,1])
            [1, 0, 1, 1] Dyck words
        """
        return f"{list(self._nu)} Dyck words"

    def _cache_key(self) -> str:
        """
        Return a cache key

        TESTS::

            sage: NuDyckWords([1,0,1,1])._cache_key()
            '1011'
        """
        return str(self._nu)

    def _an_element_(self):
        r"""
        Return an element.

        TESTS::

            sage: NuDyckWords('101').an_element()
            [1, 0, 1]
        """
        return self.element_class(self, self._nu)

    def __iter__(self, N=[], D=[], i=None, X=None):
        """
        Iterate over ``self``.

        The iterator interchanges a 0,1 pair whenever the 0 comes before a 1

        EXAMPLES::

            sage: it = NuDyckWords('101010').__iter__()
            sage: [i for i in it]
            [[1, 0, 1, 0, 1, 0],
             [1, 1, 0, 0, 1, 0],
             [1, 0, 1, 1, 0, 0],
             [1, 1, 0, 1, 0, 0],
             [1, 1, 1, 0, 0, 0]]
        """
        # Define successor function for recursion
        def transpose_close_open(N):
            for k, v in enumerate(N._list):
                if k > 0 and v == ndw_open_symbol:
                    w = N._list[k - 1]
                    if w == ndw_close_symbol:
                        new = N._list[:k - 1] + [v, w] + N._list[k + 1:]
                        yield self.element_class(self, new)

        RES = RecursivelyEnumeratedSet([self.element_class(self, self._nu)],
                                       transpose_close_open)
        return RES.breadth_first_search_iterator()

    def cardinality(self):
        r"""
        Return the number of `\nu`-Dyck words.

        EXAMPLES::

            sage: NDW = NuDyckWords('101010'); NDW.cardinality()
            5
            sage: NDW = NuDyckWords('1010010'); NDW.cardinality()
            7
            sage: NDW = NuDyckWords('100100100'); NDW.cardinality()
            12
        """
        return Integer(len([1 for _ in self.__iter__()]))


def to_word_path(word):
    r"""
    Convert input into a word path over an appropriate alphabet.

    Helper function.

    INPUT:

    - ``word`` -- word to convert to wordpath

    OUTPUT: a ``FiniteWordPath_north_east`` object

    EXAMPLES::

        sage: from sage.combinat.nu_dyck_word import to_word_path
        sage: wp = to_word_path('NEENENEN'); wp
        Path: 10010101
        sage: from sage.combinat.words.paths import FiniteWordPath_north_east
        sage: isinstance(wp,FiniteWordPath_north_east)
        True
        sage: to_word_path('1001')
        Path: 1001
        sage: to_word_path([0,1,0,0,1,0])
        Path: 010010
    """
    # If we already have the object, don't worry
    if isinstance(word, FiniteWordPath_north_east):
        return word

    # If we have a Nu Dyck Path, return the path it contains
    if isinstance(word, NuDyckWord):
        return word.path()

    # if we have a string, convert to list
    if isinstance(word, str):
        word = map(replace_dyck_char, word)

    # By default "north" is first symbol, "east" is second symbol
    P = WordPaths_north_east([ndw_open_symbol, ndw_close_symbol])

    return P(word)


def path_weakly_above_other(path, other) -> bool:
    r"""
    Test if ``path`` is weakly above ``other``.

    A path `P` is wealy above another path `Q` if `P` and `Q` are the same
    length and if any prefix of length `n` of `Q` contains more North steps
    than the prefix of length `n` of `P`.

    INPUT:

    - ``path`` -- the path to verify is weakly above the other path

    - ``other`` -- the other path to verify is weakly below the path

    OUTPUT: boolean

    EXAMPLES::

        sage: from sage.combinat.nu_dyck_word import path_weakly_above_other
        sage: path_weakly_above_other('1001','0110')
        False
        sage: path_weakly_above_other('1001','0101')
        True
        sage: path_weakly_above_other('1111','0101')
        False
        sage: path_weakly_above_other('111100','0101')
        False
    """
    # Ensure we have word paths:
    path = to_word_path(path)
    other = to_word_path(other)

    # Must be same length and must have same height
    if path.length() != other.length() or path.height() != other.height():
        return False

    # path is above other if height is always >= height of other
    p_height = path.height_vector()
    o_height = other.height_vector()
    return all(p_h >= o_h for p_h, o_h in zip(p_height, o_height))



r"""
The Real Line and Open Intervals

The class :class:`OpenInterval` implement open intervals as 1-dimensional
differentiable manifolds over `\RR`. The derived class :class:`RealLine` is
devoted to `\RR` itself, as the open interval `(-\infty, +\infty)`.

AUTHORS:

- Eric Gourgoulhon (2015): initial version
- Travis Scrimshaw (2016): review tweaks

REFERENCES:

- [Lee2013]_
"""
#*****************************************************************************
#       Copyright (C) 2015 Eric Gourgoulhon <eric.gourgoulhon@obspm.fr>
#       Copyright (C) 2015 Michal Bejger <bejger@camk.edu.pl>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.categories.manifolds import Manifolds
from sage.manifolds.differentiable.manifold import DifferentiableManifold
from sage.manifolds.structure import RealDifferentialStructure
from sage.misc.latex import latex
from sage.rings.infinity import infinity, minus_infinity
from sage.rings.real_mpfr import RR
from sage.symbolic.ring import SR
from sage.typeset.unicode_characters import unicode_mathbbR


class OpenInterval(DifferentiableManifold):
    r"""
    Open interval as a 1-dimensional differentiable manifold over `\RR`.

    INPUT:

    - ``lower`` -- lower bound of the interval (possibly ``-Infinity``)
    - ``upper`` -- upper bound of the interval (possibly ``+Infinity``)
    - ``ambient_interval`` -- (default: ``None``) another open interval,
      to which the constructed interval is a subset of
    - ``name`` -- (default: ``None``) string; name (symbol) given to
      the interval; if ``None``, the name is constructed from ``lower``
      and ``upper``
    - ``latex_name`` -- (default: ``None``) string; LaTeX symbol to denote the
      interval; if ``None``, the LaTeX symbol is constructed from ``lower``
      and ``upper`` if ``name`` is ``None``, otherwise, it is set to ``name``
    - ``coordinate`` -- (default: ``None``) string defining the symbol of the
      canonical coordinate set on the interval; if none is provided and
      ``names`` is ``None``, the symbol 't' is used
    - ``names`` -- (default: ``None``) used only when ``coordinate`` is
      ``None``: it must be a single-element tuple containing the canonical
      coordinate symbol (this is guaranteed if the shortcut ``<names>`` is
      used, see examples below)
    - ``start_index`` -- (default: 0) unique value of the index for vectors
      and forms on the interval manifold

    EXAMPLES:

    The interval `(0,\pi)`::

        sage: I = manifolds.OpenInterval(0, pi); I
        Real interval (0, pi)
        sage: latex(I)
        \left(0, \pi\right)

    ``I`` is a 1-dimensional smooth manifold over `\RR`::

        sage: I.category()
        Category of smooth connected manifolds over Real Field with 53 bits of
         precision
        sage: I.base_field()
        Real Field with 53 bits of precision
        sage: dim(I)
        1

    It is infinitely differentiable (smooth manifold)::

        sage: I.diff_degree()
        +Infinity

    The instance is unique (as long as the constructor arguments
    are the same)::

        sage: I is manifolds.OpenInterval(0, pi)
        True
        sage: I is manifolds.OpenInterval(0, pi, name='I')
        False

    The display of the interval can be customized::

        sage: I  # default display
        Real interval (0, pi)
        sage: latex(I)  # default LaTeX display
        \left(0, \pi\right)
        sage: I1 = manifolds.OpenInterval(0, pi, name='I'); I1
        Real interval I
        sage: latex(I1)
        I
        sage: I2 = manifolds.OpenInterval(0, pi, name='I', latex_name=r'\mathcal{I}'); I2
        Real interval I
        sage: latex(I2)
        \mathcal{I}

    ``I`` is endowed with a canonical chart::

        sage: I.canonical_chart()
        Chart ((0, pi), (t,))
        sage: I.canonical_chart() is I.default_chart()
        True
        sage: I.atlas()
        [Chart ((0, pi), (t,))]

    The canonical coordinate is returned by the method
    :meth:`canonical_coordinate`::

        sage: I.canonical_coordinate()
        t
        sage: t = I.canonical_coordinate()
        sage: type(t)
        <class 'sage.symbolic.expression.Expression'>

    However, it can be obtained in the same step as the interval construction
    by means of the shortcut ``I.<names>``::

        sage: I.<t> = manifolds.OpenInterval(0, pi)
        sage: t
        t
        sage: type(t)
        <class 'sage.symbolic.expression.Expression'>

    The trick is performed by the Sage preparser::

        sage: preparse("I.<t> = manifolds.OpenInterval(0, pi)")
        "I = manifolds.OpenInterval(Integer(0), pi, names=('t',)); (t,) = I._first_ngens(1)"

    In particular the shortcut can be used to set a canonical
    coordinate symbol different from ``'t'``::

        sage: J.<x> = manifolds.OpenInterval(0, pi)
        sage: J.canonical_chart()
        Chart ((0, pi), (x,))
        sage: J.canonical_coordinate()
        x

    The LaTeX symbol of the canonical coordinate can be adjusted via
    the same syntax as a chart declaration (see
    :class:`~sage.manifolds.chart.RealChart`)::

        sage: J.<x> = manifolds.OpenInterval(0, pi, coordinate=r'x:\xi')
        sage: latex(x)
        {\xi}
        sage: latex(J.canonical_chart())
        \left(\left(0, \pi\right),({\xi})\right)

    An element of the open interval ``I``::

        sage: x = I.an_element(); x
        Point on the Real interval (0, pi)
        sage: x.coord() # coordinates in the default chart = canonical chart
        (1/2*pi,)

    As for any manifold subset, a specific element of ``I`` can be created
    by providing a tuple containing its coordinate(s) in a given chart::

        sage: x = I((2,)) # (2,) = tuple of coordinates in the canonical chart
        sage: x
        Point on the Real interval (0, pi)

    But for convenience, it can also be created directly from the coordinate::

        sage: x = I(2); x
        Point on the Real interval (0, pi)
        sage: x.coord()
        (2,)
        sage: I(2) == I((2,))
        True

    By default, the coordinates passed for the element ``x`` are those
    relative to the canonical chart::

        sage: I(2) ==  I((2,), chart=I.canonical_chart())
        True

    The lower and upper bounds of the interval ``I``::

        sage: I.lower_bound()
        0
        sage: I.upper_bound()
        pi

    One of the endpoint can be infinite::

        sage: J = manifolds.OpenInterval(1, +oo); J
        Real interval (1, +Infinity)
        sage: J.an_element().coord()
        (2,)

    The construction of a subinterval can be performed via the argument
    ``ambient_interval`` of ``OpenInterval``::

        sage: J = manifolds.OpenInterval(0, 1, ambient_interval=I); J
        Real interval (0, 1)

    However, it is recommended to use the method :meth:`open_interval`
    instead::

        sage: J = I.open_interval(0, 1); J
        Real interval (0, 1)
        sage: J.is_subset(I)
        True
        sage: J.manifold() is I
        True

    A subinterval of a subinterval::

        sage: K = J.open_interval(1/2, 1); K
        Real interval (1/2, 1)
        sage: K.is_subset(J)
        True
        sage: K.is_subset(I)
        True
        sage: K.manifold() is I
        True

    We have::

        sage: list(I.subset_family())
        [Real interval (0, 1), Real interval (0, pi), Real interval (1/2, 1)]
        sage: list(J.subset_family())
        [Real interval (0, 1), Real interval (1/2, 1)]
        sage: list(K.subset_family())
        [Real interval (1/2, 1)]

    As any open subset of a manifold, open subintervals are created in a
    category of subobjects of smooth manifolds::

        sage: J.category()
        Join of Category of subobjects of sets and Category of smooth manifolds
         over Real Field with 53 bits of precision and Category of connected
         manifolds over Real Field with 53 bits of precision
        sage: K.category()
        Join of Category of subobjects of sets and Category of smooth manifolds
         over Real Field with 53 bits of precision and Category of connected
         manifolds over Real Field with 53 bits of precision

    On the contrary, ``I``, which has not been created as a subinterval,
    is in the category of smooth manifolds (see
    :class:`~sage.categories.manifolds.Manifolds`)::

        sage: I.category()
        Category of smooth connected manifolds over Real Field with 53 bits of
         precision

    and we have::

        sage: J.category() is I.category().Subobjects()
        True

    All intervals are parents::

        sage: x = J(1/2); x
        Point on the Real interval (0, pi)
        sage: x.parent() is J
        True
        sage: y = K(3/4); y
        Point on the Real interval (0, pi)
        sage: y.parent() is K
        True

    We have::

        sage: x in I, x in J, x in K
        (True, True, False)
        sage: y in I, y in J, y in K
        (True, True, True)

    The canonical chart of subintervals is inherited from the canonical chart
    of the parent interval::

        sage: XI = I.canonical_chart(); XI
        Chart ((0, pi), (t,))
        sage: XI.coord_range()
        t: (0, pi)
        sage: XJ = J.canonical_chart(); XJ
        Chart ((0, 1), (t,))
        sage: XJ.coord_range()
        t: (0, 1)
        sage: XK = K.canonical_chart(); XK
        Chart ((1/2, 1), (t,))
        sage: XK.coord_range()
        t: (1/2, 1)
    """
    @staticmethod
    def __classcall_private__(cls, lower, upper, ambient_interval=None,
                              name=None, latex_name=None, coordinate=None,
                              names=None, start_index=0):
        r"""
        Determine the correct interval to return based upon the input.

        TESTS:

        Check whether :issue:`30830` is fixed::

            sage: I = manifolds.OpenInterval(0,2)
            sage: J = manifolds.OpenInterval(0,1, ambient_interval=I, coordinate='t')
            sage: I.open_interval(0,1)
            Real interval (0, 1)
        """
        if ambient_interval:
            # cope the UniqueRepresentation framework for subintervals and
            # reset irrelevant information only:
            coordinate = None
            names = None
            start_index = 0
        return super().__classcall__(cls, lower, upper,
                                     ambient_interval=ambient_interval, name=name,
                                     latex_name=latex_name, coordinate=coordinate,
                                     names=names, start_index=start_index)

    def __init__(self, lower, upper, ambient_interval=None,
                 name=None, latex_name=None,
                 coordinate=None, names=None, start_index=0):
        r"""
        Construct an open interval.

        TESTS::

            sage: I = manifolds.OpenInterval(-1,1); I
            Real interval (-1, 1)
            sage: TestSuite(I).run(skip='_test_elements')  # pickling of elements fails
            sage: J = manifolds.OpenInterval(-oo, 2); J
            Real interval (-Infinity, 2)
            sage: TestSuite(J).run(skip='_test_elements')  # pickling of elements fails
        """
        if latex_name is None:
            if name is None:
                latex_name = r"\left({}, {}\right)".format(latex(lower), latex(upper))
            else:
                latex_name = name
        if name is None:
            name = "({}, {})".format(lower, upper)
        if ambient_interval is None:
            ambient_manifold = None
        else:
            if not isinstance(ambient_interval, OpenInterval):
                raise TypeError("the argument ambient_interval must be an open interval")
            ambient_manifold = ambient_interval.manifold()
        field = 'real'
        structure = RealDifferentialStructure()
        category = Manifolds(RR).Smooth().Connected()
        DifferentiableManifold.__init__(self, 1, name, field, structure,
                                        base_manifold=ambient_manifold,
                                        latex_name=latex_name,
                                        start_index=start_index,
                                        category=category)
        if ambient_interval is None:
            if coordinate is None:
                if names is None:
                    coordinate = 't'
                else:
                    coordinate = names[0]
        else:
            if lower < ambient_interval.lower_bound():
                raise ValueError("the lower bound is smaller than that of "
                                 + "the containing interval")
            if upper > ambient_interval.upper_bound():
                raise ValueError("the upper bound is larger than that of "
                                 + "the containing interval")
            self.declare_subset(ambient_interval)
            ambient_interval._top_subsets.add(self)
        if lower != minus_infinity:
            if upper != infinity:
                restrictions = lambda t: [t > lower, t < upper]
            else:
                restrictions = lambda t: t > lower
        else:
            if upper != infinity:
                restrictions = lambda t: t < upper
            else:
                restrictions = None
        if ambient_interval is None:
            self._canon_chart = self.chart(coordinates=coordinate,
                                           coord_restrictions=restrictions)
        else:
            self._canon_chart = ambient_interval.canonical_chart().restrict(self,
                                                                            restrictions=restrictions)
        self._lower = lower
        self._upper = upper

    def _repr_(self):
        r"""
        String representation of ``self``.

        TESTS::

            sage: I = manifolds.OpenInterval(-1, pi)
            sage: I
            Real interval (-1, pi)
            sage: I = manifolds.OpenInterval(-1, +oo)
            sage: I
            Real interval (-1, +Infinity)
            sage: I = manifolds.OpenInterval(-oo,0)
            sage: I
            Real interval (-Infinity, 0)
        """
        return "Real interval " + self._name

    def _first_ngens(self, n):
        r"""
        Return the coordinate of the canonical chart.

        This is useful only for the use of Sage preparser.

        INPUT:

        - ``n`` -- the number of coordinates: must be 1

        TESTS::

            sage: I = manifolds.OpenInterval(-1, 1)
            sage: I._first_ngens(1)
            (t,)
            sage: I = manifolds.OpenInterval(-1, 1, coordinate='x')
            sage: I._first_ngens(1)
            (x,)
            sage: I = manifolds.OpenInterval(-1, 1, names=('x',))
            sage: I._first_ngens(1)
            (x,)
        """
        return self._canon_chart[:]

    def _element_constructor_(self, coords=None, chart=None, name=None,
                              latex_name=None, check_coords=True):
        r"""
        Construct an element of ``self``.

        This is a redefinition of
        :meth:`sage.manifolds.differentiable.DifferentiableManifold._element_constructor_`
        to allow for construction from a number (considered as the canonical
        coordinate).

        INPUT:

        - ``coords`` -- (default: ``None``) either (i) the point coordinates
          (as a single-element tuple or list) in the chart ``chart``, (ii) the
          value of the point coordinate in the chart ``chart``, or (iii)
          another point in the interval
        - ``chart`` -- (default: ``None``) chart in which the coordinates are
          given; if none is provided, the coordinates are assumed to refer to
          the interval's default chart
        - ``name`` -- (default: ``None``) name given to the point
        - ``latex_name`` -- (default: ``None``) LaTeX symbol to denote the
          point; if none is provided, the LaTeX symbol is set to ``name``
        - ``check_coords`` -- boolean (default: ``True``); determines whether
          ``coords`` are valid coordinates for the chart ``chart``. For symbolic
          coordinates, it is recommended to set ``check_coords`` to ``False``.

        OUTPUT: :class:`~sage.manifolds.point.TopologicalManifoldPoint`
        representing a point in the current interval

        EXAMPLES::

            sage: I = manifolds.OpenInterval(-1, 4)
            sage: I((2,))  # standard used of TopologicalManifoldSubset._element_constructor_
            Point on the Real interval (-1, 4)
            sage: I(2)  # specific use with a single coordinate
            Point on the Real interval (-1, 4)
            sage: I(2).coord()
            (2,)
            sage: I(2) == I((2,))
            True
            sage: I(pi)
            Point on the Real interval (-1, 4)
            sage: I(pi).coord()
            (pi,)
            sage: I(8)
            Traceback (most recent call last):
            ...
            ValueError: the coordinates (8,) are not valid on the Chart
             ((-1, 4), (t,))
        """
        if coords in SR:
            coords = (coords,)
        return super()._element_constructor_(coords=coords,
                                             chart=chart, name=name, latex_name=latex_name,
                                             check_coords=check_coords)

    def _Hom_(self, other, category=None):
        r"""
        Construct the set of curves in ``other`` with parameter in ``self``.

        INPUT:

        - ``other`` -- a differentiable manifold `M`
        - ``category`` -- (default: ``None``) not used here (to ensure
          compatibility with generic hook ``_Hom_``)

        OUTPUT: the set of curves `I \to M`,  where `I` is ``self``

        .. SEEALSO::

            :class:`~sage.manifolds.differentiable.manifold_homset.DifferentiableCurveSet`
            for more documentation.

        TESTS::

            sage: I = manifolds.OpenInterval(-1,1)
            sage: M = Manifold(3, 'M')
            sage: H = I._Hom_(M); H
            Set of Morphisms from Real interval (-1, 1) to 3-dimensional
             differentiable manifold M in Category of smooth manifolds over Real
             Field with 53 bits of precision
            sage: H is Hom(I, M)
            True
        """
        from sage.manifolds.differentiable.manifold_homset import DifferentiableCurveSet
        return DifferentiableCurveSet(self, other)

    def canonical_chart(self):
        r"""
        Return the canonical chart defined on ``self``.

        OUTPUT: :class:`~sage.manifolds.differentiable.chart.RealDiffChart`

        EXAMPLES:

        Canonical chart on the interval `(0, \pi)`::

            sage: I = manifolds.OpenInterval(0, pi)
            sage: I.canonical_chart()
            Chart ((0, pi), (t,))
            sage: I.canonical_chart().coord_range()
            t: (0, pi)

        The symbol used for the coordinate of the canonical chart is that
        defined during the construction of the interval::

            sage: I.<x> = manifolds.OpenInterval(0, pi)
            sage: I.canonical_chart()
            Chart ((0, pi), (x,))
        """
        return self._canon_chart

    def canonical_coordinate(self):
        r"""
        Return the canonical coordinate defined on the interval.

        OUTPUT: the symbolic variable representing the canonical coordinate

        EXAMPLES:

        Canonical coordinate on the interval `(0, \pi)`::

            sage: I = manifolds.OpenInterval(0, pi)
            sage: I.canonical_coordinate()
            t
            sage: type(I.canonical_coordinate())
            <class 'sage.symbolic.expression.Expression'>
            sage: I.canonical_coordinate().is_real()
            True

        The canonical coordinate is the first (unique) coordinate of the
        canonical chart::

            sage: I.canonical_coordinate() is I.canonical_chart()[0]
            True

        Its default symbol is `t`; but it can be customized during the
        creation of the interval::

            sage: I = manifolds.OpenInterval(0, pi, coordinate='x')
            sage: I.canonical_coordinate()
            x
            sage: I.<x> = manifolds.OpenInterval(0, pi)
            sage: I.canonical_coordinate()
            x
        """
        return self._canon_chart._xx[0]

    def lower_bound(self):
        r"""
        Return the lower bound (infimum) of the interval.

        EXAMPLES::

            sage: I = manifolds.OpenInterval(1/4, 3)
            sage: I.lower_bound()
            1/4
            sage: J = manifolds.OpenInterval(-oo, 2)
            sage: J.lower_bound()
            -Infinity

        An alias of :meth:`lower_bound` is :meth:`inf`::

            sage: I.inf()
            1/4
            sage: J.inf()
            -Infinity
        """
        return self._lower

    inf = lower_bound

    def upper_bound(self):
        r"""
        Return the upper bound (supremum) of the interval.

        EXAMPLES::

            sage: I = manifolds.OpenInterval(1/4, 3)
            sage: I.upper_bound()
            3
            sage: J = manifolds.OpenInterval(1, +oo)
            sage: J.upper_bound()
            +Infinity

        An alias of :meth:`upper_bound` is :meth:`sup`::

            sage: I.sup()
            3
            sage: J.sup()
            +Infinity
        """
        return self._upper

    sup = upper_bound

    def open_interval(self, lower, upper, name=None, latex_name=None):
        r"""
        Define an open subinterval of ``self``.

        INPUT:

        - ``lower`` -- lower bound of the subinterval (possibly ``-Infinity``)
        - ``upper`` -- upper bound of the subinterval (possibly ``+Infinity``)
        - ``name`` -- (default: ``None``) string; name (symbol) given to the
          subinterval; if ``None``, the name is constructed from ``lower`` and
          ``upper``
        - ``latex_name`` -- (default: ``None``) string; LaTeX symbol to denote
          the subinterval; if ``None``, the LaTeX symbol is constructed from
          ``lower`` and ``upper`` if ``name`` is ``None``, otherwise, it is set
          to ``name``

        OUTPUT:

        - :class:`~sage.manifolds.differentiable.examples.real_line.OpenInterval`
          representing the open interval (``lower``, ``upper``)

        EXAMPLES:

        The interval `(0, \pi)` as a subinterval of `(-4, 4)`::

            sage: I = manifolds.OpenInterval(-4, 4)
            sage: J = I.open_interval(0, pi); J
            Real interval (0, pi)
            sage: J.is_subset(I)
            True
            sage: list(I.subset_family())
            [Real interval (-4, 4), Real interval (0, pi)]

        ``J`` is considered as an open submanifold of ``I``::

            sage: J.manifold() is I
            True

        The subinterval `(-4, 4)` is ``I`` itself::

            sage: I.open_interval(-4, 4) is I
            True
        """
        if lower == self._lower and upper == self._upper:
            return self
        # To cope with the unique representation framework, we have to
        # distinguish several cases, instead of performing a mere
        # return OpenInterval(lower, upper, ambient_interval=self, name=name,
        #                     latex_name=latex_name)
        if name is None:
            if latex_name is None:
                return OpenInterval(lower, upper, ambient_interval=self)
            return OpenInterval(lower, upper, ambient_interval=self,
                                latex_name=latex_name)
        if latex_name is None:
            return OpenInterval(lower, upper, ambient_interval=self, name=name)
        return OpenInterval(lower, upper, ambient_interval=self, name=name,
                            latex_name=latex_name)


#******************************************************************************

class RealLine(OpenInterval):
    r"""
    Field of real numbers, as a differentiable manifold of dimension 1 (real
    line) with a canonical coordinate chart.

    INPUT:

    - ``name`` -- (default: ``'R'``) string; name (symbol) given to
      the real line
    - ``latex_name`` -- (default: ``r'\Bold{R}'``) string; LaTeX symbol to
      denote the real line
    - ``coordinate`` -- (default: ``None``) string defining the symbol of the
      canonical coordinate set on the real line; if none is provided and
      ``names`` is ``None``, the symbol 't' is used
    - ``names`` -- (default: ``None``) used only when ``coordinate`` is
      ``None``: it must be a single-element tuple containing the canonical
      coordinate symbol (this is guaranteed if the shortcut ``<names>`` is
      used, see examples below)
    - ``start_index`` -- (default: 0) unique value of the index for vectors
      and forms on the real line manifold

    EXAMPLES:

    Constructing the real line without any argument::

        sage: R = manifolds.RealLine() ; R
        Real number line ℝ
        sage: latex(R)
        \Bold{R}

    ``R`` is a 1-dimensional real smooth manifold::

        sage: R.category()
        Category of smooth connected manifolds over Real Field with 53 bits of
         precision
        sage: isinstance(R, sage.manifolds.differentiable.manifold.DifferentiableManifold)
        True
        sage: dim(R)
        1

    It is endowed with a canonical chart::

        sage: R.canonical_chart()
        Chart (ℝ, (t,))
        sage: R.canonical_chart() is R.default_chart()
        True
        sage: R.atlas()
        [Chart (ℝ, (t,))]

    The instance is unique (as long as the constructor arguments are the
    same)::

        sage: R is manifolds.RealLine()
        True
        sage: R is manifolds.RealLine(latex_name='R')
        False

    The canonical coordinate is returned by the method
    :meth:`~sage.manifolds.differentiable.examples.real_line.OpenInterval.canonical_coordinate`::

        sage: R.canonical_coordinate()
        t
        sage: t = R.canonical_coordinate()
        sage: type(t)
        <class 'sage.symbolic.expression.Expression'>

    However, it can be obtained in the same step as the real line construction
    by means of the shortcut ``R.<names>``::

        sage: R.<t> = manifolds.RealLine()
        sage: t
        t
        sage: type(t)
        <class 'sage.symbolic.expression.Expression'>

    The trick is performed by Sage preparser::

        sage: preparse("R.<t> = manifolds.RealLine()")
        "R = manifolds.RealLine(names=('t',)); (t,) = R._first_ngens(1)"

    In particular the shortcut is to be used to set a canonical
    coordinate symbol different from 't'::

        sage: R.<x> = manifolds.RealLine()
        sage: R.canonical_chart()
        Chart (ℝ, (x,))
        sage: R.atlas()
        [Chart (ℝ, (x,))]
        sage: R.canonical_coordinate()
        x

    The LaTeX symbol of the canonical coordinate can be adjusted via the same
    syntax as a chart declaration (see
    :class:`~sage.manifolds.chart.RealChart`)::

        sage: R.<x> = manifolds.RealLine(coordinate=r'x:\xi')
        sage: latex(x)
        {\xi}
        sage: latex(R.canonical_chart())
        \left(\Bold{R},({\xi})\right)

    The LaTeX symbol of the real line itself can also be customized::

        sage: R.<x> = manifolds.RealLine(latex_name=r'\mathbb{R}')
        sage: latex(R)
        \mathbb{R}

    Elements of the real line can be constructed directly from a number::

        sage: p = R(2) ; p
        Point on the Real number line ℝ
        sage: p.coord()
        (2,)
        sage: p = R(1.742) ; p
        Point on the Real number line ℝ
        sage: p.coord()
        (1.74200000000000,)

    Symbolic variables can also be used::

        sage: p = R(pi, name='pi') ; p
        Point pi on the Real number line ℝ
        sage: p.coord()
        (pi,)
        sage: a = var('a')
        sage: p = R(a) ; p
        Point on the Real number line ℝ
        sage: p.coord()
        (a,)

    The real line is considered as the open interval `(-\infty, +\infty)`::

        sage: isinstance(R, sage.manifolds.differentiable.examples.real_line.OpenInterval)
        True
        sage: R.lower_bound()
        -Infinity
        sage: R.upper_bound()
        +Infinity

    A real interval can be created from ``R`` means of the method
    :meth:`~sage.manifolds.differentiable.examples.real_line.OpenInterval.open_interval`::

        sage: I = R.open_interval(0, 1); I
        Real interval (0, 1)
        sage: I.manifold()
        Real number line ℝ
        sage: list(R.subset_family())
        [Real interval (0, 1), Real number line ℝ]
    """
    @staticmethod
    def __classcall__(cls, name=unicode_mathbbR, latex_name=r'\Bold{R}',
                      coordinate=None, names=None, start_index=0):
        r"""
        Determine the correct interval to return based upon the input.

        TESTS::

            sage: R = manifolds.RealLine(); R
            Real number line ℝ
            sage: R1 = manifolds.RealLine('ℝ'); R1
            Real number line ℝ
            sage: R is R1
            True
        """
        return super().__classcall__(cls, name=name,
                                     latex_name=latex_name,
                                     coordinate=coordinate,
                                     names=names, start_index=start_index)

    def __init__(self, name=unicode_mathbbR, latex_name=r'\Bold{R}',
                 coordinate=None, names=None, start_index=0):
        r"""
        Construct the real line manifold.

        TESTS::

            sage: R = manifolds.RealLine() ; R
            Real number line ℝ
            sage: R.category()
            Category of smooth connected manifolds over Real Field with 53 bits
             of precision
            sage: TestSuite(R).run(skip='_test_elements')  # pickling of elements fails
        """
        OpenInterval.__init__(self, minus_infinity, infinity, name=name,
                              latex_name=latex_name, coordinate=coordinate,
                              names=names, start_index=start_index)

    def _repr_(self):
        r"""
        String representation of ``self``.

        TESTS::

            sage: R = manifolds.RealLine()
            sage: R._repr_()
            'Real number line ℝ'
            sage: R = manifolds.RealLine(name='r')
            sage: R._repr_()
            'Real number line r'
        """
        return "Real number line " + self._name
